/*
 * File:   UserInterface.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */
 
 package gui;
 
import gui.view.*;
import gui.context.*;
import gui.graph.*;
import sim.Simulation;
import sim.livestock.*;
import main.PastureSim;

import java.awt.*;
import java.awt.Color;
import java.awt.Container;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.*;
import javax.swing.*;
import javax.swing.SpringLayout;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JButton;

/** The main user interface. All other UI components stem from this */
public class UserInterface extends JFrame{

	public static final int INIT_WIDTH = 400;
	public static final int INIT_HEIGHT = 400;
	public static final int CONTEXT_HEIGHT = 45;
	
	public static Image iconImage = Toolkit.getDefaultToolkit().getImage(UserInterface.class.getResource("icon/icon.gif"));
	public static ImageIcon icon = new ImageIcon(iconImage);
	
	private Simulation sim; // A reference to the simulation
	private Container mainPanel;
	private SpringLayout layout;
	private ActionPanel view;
	private ContextPanel context;
	private MenuBar menu;
	private boolean loaded;
	
	private MediaTracker mediaTracker;
	private GraphicsConfiguration gc;
	private ImageProgressDialog imgDialog;
	
	
	public GraphWindow nutritionWindow;
	public GraphWindow pastureWindow;
	public GraphWindow behaviourWindow;
	
		
	public UserInterface(){
		super("PastureSim");
		this.sim = null;
		this.loaded = false;
		
		setResizable(false);
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

		WindowListener closeAction = new WindowListener(){
			public void windowActivated(WindowEvent e){}
			public void windowClosed(WindowEvent e){}
			public void windowClosing(WindowEvent e){ close(); }
			public void windowDeactivated(WindowEvent e){}
			public void windowDeiconified(WindowEvent e){}
			public void windowIconified(WindowEvent e){}
			public void windowOpened(WindowEvent e){}
		}; addWindowListener(closeAction);
		
		//Schedule a job for the event-dispatching thread:
		//creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			public void run() { generate(); }
		});
		
	}
	
	/** The constructor sets the fields and dispatches a new
	  * thread to display the user interface */
	public UserInterface(File simulationFile, File f, int chosenArea, int chosenRate) throws Exception{
		super("PastureSim");
		newInterface(new Simulation(simulationFile, f, chosenArea, chosenRate, this));
	}public UserInterface(File simulationFile) throws Exception{
		super("PastureSim");
		newInterface(new Simulation(simulationFile, this));
	}private void newInterface(Simulation sim){
		this.sim = sim;
		this.loaded = true;
		setResizable(false);
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		WindowListener closeAction = new WindowListener(){
			public void windowActivated(WindowEvent e){}
			public void windowClosed(WindowEvent e){}
			public void windowClosing(WindowEvent e){ close(); }
			public void windowDeactivated(WindowEvent e){}
			public void windowDeiconified(WindowEvent e){}
			public void windowIconified(WindowEvent e){}
			public void windowOpened(WindowEvent e){}
		}; addWindowListener(closeAction);
		
		//Schedule a job for the event-dispatching thread:
		//creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			public void run() { generate(); }
		});
	}
	
	public void loadSimulation(File simulationFile, File f, int chosenArea, int chosenRate) throws Exception{
		loadSimulation(new Simulation(simulationFile, f, chosenArea, chosenRate, this));
	}public void loadSimulation(File simulationFile) throws Exception{
		loadSimulation(new Simulation(simulationFile, this));
	}private void loadSimulation(Simulation sim){
		this.sim = sim;
		this.loaded = true;
		setSize(sim.getPastureSize().width+10, sim.getPastureSize().height+80);
		context.setSimulation(sim);
		view.setSimulation(sim);
		menu.setSimulation(sim);
		switchContext();
		menu.enableItems();
		context.enableButtons();
		pack();
		repaint();
	}
	
	public void refreshMenuItems(){ menu.enableItems(); menu.repaint(); }
	public void refreshContextItems(){
		context.enableButtons();
		context.repaint();
	}
	public void refreshMenu(){ menu.repaint(); }
	
	public void refreshGraphs(){
		if(nutritionWindow != null) nutritionWindow.repaint();
		if(pastureWindow != null) pastureWindow.repaint();
		if(behaviourWindow != null) behaviourWindow.repaint();
	}
	
	public boolean isSimulationLoaded(){ return loaded; }

	
	/** Generate the gui */
	private void generate(){
	
		nutritionWindow = null;
		pastureWindow = null;
		behaviourWindow = null;
	
		mediaTracker = new MediaTracker(this);
		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice gs = ge.getDefaultScreenDevice();
		gc = gs.getDefaultConfiguration();

		setIconImage(iconImage);
   
		// Set up the content pane
		mainPanel = getContentPane();
		layout = new SpringLayout();
		mainPanel.setLayout(layout);

		view = new ReadOnlyActionPanel(sim, this);
		mainPanel.add(view);
	
		// Add the menu bar
		menu = new MenuBar(sim, this); mainPanel.add(menu);
		
		// Add the context pane
		if(sim == null) context = new EmptyContextPanel(sim);
		else context = new SimulationContextPanel(sim);
		mainPanel.add(context);
	
				
		layout.putConstraint(SpringLayout.NORTH, menu,      0,              SpringLayout.NORTH, mainPanel);
		layout.putConstraint(SpringLayout.WEST,  view,      0,              SpringLayout.WEST,  mainPanel);
		layout.putConstraint(SpringLayout.EAST,  mainPanel, 0,              SpringLayout.EAST,  view);
		layout.putConstraint(SpringLayout.SOUTH, mainPanel, CONTEXT_HEIGHT, SpringLayout.SOUTH, view);
		
		layout.putConstraint(SpringLayout.NORTH, view,      20,             SpringLayout.NORTH, mainPanel);
		layout.putConstraint(SpringLayout.SOUTH, context,   0,              SpringLayout.SOUTH, mainPanel);
		
		layout.putConstraint(SpringLayout.WEST,  menu,      0,              SpringLayout.WEST, mainPanel);
		layout.putConstraint(SpringLayout.WEST,  context,   0,              SpringLayout.WEST, mainPanel);
				
		pack();
		setVisible(!main.PastureSim.textOnly);
	}
	
	
	// Alters the user interface components when we enter a new mode
	public void switchContext(){
		
		if(sim.getMode() != Simulation.MODE_HISTORY)
			sim.gotoState(sim.STATE_CURRENT);

			
		if(view instanceof PastureActionPanel){
			((PastureActionPanel)view).notifyExit();
		}
	
		mainPanel.remove(view);
		mainPanel.remove(context);
		
		switch(sim.getMode()){
		case Simulation.MODE_ADJUST:
			context.disableButtons();
			menu.disableSimulationItems();
			ParameterAdjustDialog d = new ParameterAdjustDialog(this);			
			d.show(sim);
			break;
		case Simulation.MODE_SIMULATE:
			if(!(view instanceof ReadOnlyActionPanel)){
				view    = new ReadOnlyActionPanel(sim, this);
			}
			context = new SimulationContextPanel(sim);
			menu.enableItems();
			break;
		case Simulation.MODE_PASTURE:
			if(!(view instanceof PastureActionPanel)){
				view    = new PastureActionPanel(sim);
			}
			if(!(context instanceof PastureContextPanel)){
				context = new PastureContextPanel(sim);
			}
			context.enableButtons();
			break;
		case Simulation.MODE_PASTURE_ADD_FIELD:
			view    = new FieldAddActionPanel(sim);
			context.disableButtons();
			break;
		case Simulation.MODE_PASTURE_REMOVE_FIELD:
			view    = new FieldRemoveActionPanel(sim);
			context.disableButtons();
			break;
		case Simulation.MODE_PASTURE_ADD_GATE:
			view    = new GateAddActionPanel(sim);
			context.disableButtons();
			break;
		case Simulation.MODE_PASTURE_REMOVE_GATE:
			view    = new GateRemoveActionPanel(sim);
			context.disableButtons();
			break;
		case Simulation.MODE_PASTURE_ADD_WATER:
			view    = new WaterAddActionPanel(sim);
			context.disableButtons();
			break;
		case Simulation.MODE_PASTURE_REMOVE_WATER:
			view    = new WaterRemoveActionPanel(sim);
			context.disableButtons();
			break;
		case Simulation.MODE_LIVESTOCK:
			if(!(view instanceof ReadOnlyActionPanel)){
				view    = new ReadOnlyActionPanel(sim, this);
			}
			if(!(context instanceof LivestockContextPanel)){
				context = new LivestockContextPanel(sim);
			}else{ context.enableButtons(); }
			break;			
		case Simulation.MODE_LIVESTOCK_ADD:
			mainPanel.add(view);
			mainPanel.add(context);
			context.disableButtons();
			LivestockAddDialog addDialog = new LivestockAddDialog(this);
			LivestockGroup gAdd = addDialog.show(sim);
			if(gAdd == null){ context.enableButtons(); sim.mode = Simulation.MODE_LIVESTOCK; }
			else{
				mainPanel.remove(view);
				view = new LivestockAddActionPanel(sim, gAdd);
			}
			break;
		case Simulation.MODE_LIVESTOCK_PLACE:
			mainPanel.add(view);
			mainPanel.add(context);
			context.disableButtons();
			LivestockSelectDialog placeDialog = new LivestockSelectDialog(this);
			LivestockGroup gPlace = placeDialog.show(LivestockSelectDialog.PLACE, sim);
			if(gPlace == null){ context.enableButtons(); sim.mode = Simulation.MODE_LIVESTOCK; }
			else{
				mainPanel.remove(view);
				view = new LivestockPlaceActionPanel(sim, gPlace);
			}
			break;
		case Simulation.MODE_LIVESTOCK_REMOVE:
			mainPanel.add(view);
			mainPanel.add(context);
			context.disableButtons();
			LivestockSelectDialog removeDialog = new LivestockSelectDialog(this);
			LivestockGroup gRemove = removeDialog.show(LivestockSelectDialog.REMOVE, sim);
			mainPanel.remove(view);
			sim.removeLivestock(gRemove);
			sim.mode = Simulation.MODE_LIVESTOCK;
			context.enableButtons();
			break;
		case Simulation.MODE_HISTORY:
			if(!(view instanceof ReadOnlyActionPanel)){
				view    = new ReadOnlyActionPanel(sim, this);
			}
			context = new HistoryContextPanel(sim);
			break;			
		case Simulation.MODE_GRAPH:
			if(!(view instanceof ReadOnlyActionPanel)){
				view    = new ReadOnlyActionPanel(sim, this);
			}
			context = new GraphContextPanel(sim);
			break;			
		default:
			if(!(view instanceof ReadOnlyActionPanel)){
				view    = new ReadOnlyActionPanel(sim, this);
			}
			if(!(context instanceof EmptyContextPanel)){
				context = new EmptyContextPanel(sim);
			}
		}
		
		mainPanel.add(view);
		mainPanel.add(context);
		
		view.revalidate();
		context.revalidate();
		
		//layout.putConstraint(SpringLayout.NORTH, menu, 0,                 SpringLayout.NORTH, mainPanel);
		layout.putConstraint(SpringLayout.WEST,  view, 0,       SpringLayout.WEST,  mainPanel);
		layout.putConstraint(SpringLayout.EAST,  mainPanel, 0,  SpringLayout.EAST,  view);
		layout.putConstraint(SpringLayout.SOUTH, mainPanel, CONTEXT_HEIGHT, SpringLayout.SOUTH, view);
		
		layout.putConstraint(SpringLayout.NORTH, view,    20, SpringLayout.NORTH, mainPanel);
		layout.putConstraint(SpringLayout.SOUTH, context, 0, SpringLayout.SOUTH, mainPanel);
		
		//layout.putConstraint(SpringLayout.WEST, menu,    0, SpringLayout.WEST, mainPanel);
		layout.putConstraint(SpringLayout.WEST, context, 0, SpringLayout.WEST, mainPanel);

	
		validate();
		
		repaint();
	}
	
	public void notifyModeSwitch(){
		if(view instanceof GateRemoveActionPanel){
			((PastureActionPanel)view).notifyExit();
		}
	}
	

	public BufferedImage loadImage(File imageFile, int id) throws IOException{

		
		if(imageFile == null) throw new IOException("Can not load image 'null'.");
		MediaTracker mediaTracker = new MediaTracker(this);
		Toolkit toolkit = Toolkit.getDefaultToolkit();
		Image image = toolkit.getImage(imageFile.toString());
		mediaTracker.addImage(image, id);
		try{ mediaTracker.waitForID(id);
		}catch (InterruptedException ie){ throw new IOException("Image loading interrupted."); }
		

		if(gc == null){
			GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
			GraphicsDevice gs = ge.getDefaultScreenDevice();
			gc = gs.getDefaultConfiguration();
		}
		
		BufferedImage bImage = gc.createCompatibleImage(image.getWidth(null), image.getHeight(null), Transparency.OPAQUE);
		//BufferedImage bImage = new BufferedImage(image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_INT_ARGB);
		//BufferedImage bImage = javax.imageio.ImageIO.read(imageFile);
		
		Graphics g = bImage.createGraphics();
		g.drawImage(image, 0, 0, null);
		g.dispose();

		
		return bImage;
	}
	
	/** Open a graph window for animal nutrition */
	public void graphNutrition(){
		nutritionWindow = (GraphWindow)new NutritionGraphWindow(this, sim);
		menu.enableItems();
	}
	/** Open a graph window for pasture state */
	public void graphPasture(){
		pastureWindow = (GraphWindow)new PastureGraphWindow(this, sim);
		menu.enableItems();
	}
	/** Open a graph window for animal behaviour */
	public void graphBehaviour(){
		behaviourWindow = (GraphWindow)new BehaviourGraphWindow(this, sim);
		menu.enableItems();
	}
	/** Allow the user interface to respond when a graph window is closed */
	public void notifyClosed(GraphWindow g){
		if(g == nutritionWindow) nutritionWindow = null;
		if(g == pastureWindow) pastureWindow = null;
		if(g == behaviourWindow) behaviourWindow = null;
		menu.enableItems();
	}
	/** Update the graphs for real time simulation */
	public void updateGraphs(){
		if(nutritionWindow != null){ nutritionWindow.update(); }
		if(pastureWindow != null){ pastureWindow.update(); }
		if(behaviourWindow != null){ behaviourWindow.update(); }
	}
	
		
	public void close(){
		if(sim != null) sim.close();
		else{ dispose(); PastureSim.cleanUp(this); }
	}

}
