/*
 * File:   ParameterAdjustDialog.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */

 
package gui.context;

import gui.Popup;
import sim.*;
import sim.livestock.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/** The dialog box for adjusting pasture parameters */
public class ParameterAdjustDialog extends JDialog implements ActionListener{

	private Simulation sim;
	private JFrame parent;
	private JButton doButton;
	private JButton cancelButton;
	
	private JTextField rainValue;
	private JTextField tempValue;
	private JTextField photoValue;
	private JTextField transValue;
	
	private double rainfall;
	private double temperature;
	private double radiation;
	private double transpiration;
	
	
	public ParameterAdjustDialog(JFrame parent){ super(parent, false); this.parent = parent; }
	
	
	/** Show the dialog */
	public void show(Simulation sim){
		this.sim = sim;
		setTitle("Environment Settings");
		setDefaultCloseOperation(DISPOSE_ON_CLOSE);
		setResizable(false);
		
		Container pane = getContentPane();
		SpringLayout layout = new SpringLayout();
		pane.setLayout(layout);
				
		JLabel rainLabel = new JLabel("Daily Rainfall [0mm - 500mm]"); pane.add(rainLabel);
		rainValue = new JTextField(""+sim.getRainfall(), 10); pane.add(rainValue);
		JLabel tempLabel = new JLabel("Temperature [0\u00B0C - 50\u00B0C ]"); pane.add(tempLabel);
		tempValue = new JTextField(""+sim.getTemperature(), 10); pane.add(tempValue);
		JLabel photoLabel = new JLabel("Photosynthetic Radiation [0MJ/m\u00B2 - 60MJ/m\u00B2]"); pane.add(photoLabel);
		photoValue = new JTextField(""+sim.getRadiation(), 10); pane.add(photoValue);
		JLabel transLabel = new JLabel("Potential Evapo-Transpiration [0mm - 250mm]"); pane.add(transLabel);
		transValue = new JTextField(""+sim.getTranspiration(), 10); pane.add(transValue);
		
		doButton = new JButton("OK");
		pane.add(doButton); doButton.addActionListener(this);
		cancelButton = new JButton("Cancel");
		pane.add(cancelButton); cancelButton.addActionListener(this);
		
		layout.putConstraint(SpringLayout.NORTH, rainLabel, 10, SpringLayout.NORTH, pane);
		layout.putConstraint(SpringLayout.WEST,  rainLabel, 10, SpringLayout.WEST, pane);
		layout.putConstraint(SpringLayout.EAST,  pane, 10, SpringLayout.EAST, transValue);
		layout.putConstraint(SpringLayout.SOUTH, pane, 10, SpringLayout.SOUTH, cancelButton);
		
		layout.putConstraint(SpringLayout.NORTH, rainValue, -2, SpringLayout.NORTH, rainLabel);
		layout.putConstraint(SpringLayout.WEST,  rainValue, 10, SpringLayout.EAST,  transLabel);
		
		layout.putConstraint(SpringLayout.NORTH, tempLabel, 20, SpringLayout.SOUTH, rainLabel);
		layout.putConstraint(SpringLayout.WEST,  tempLabel, 0,  SpringLayout.WEST,  rainLabel);
		layout.putConstraint(SpringLayout.NORTH, tempValue, -2, SpringLayout.NORTH, tempLabel);
		layout.putConstraint(SpringLayout.WEST,  tempValue, 10, SpringLayout.EAST,  transLabel);
		
		layout.putConstraint(SpringLayout.NORTH, photoLabel, 20, SpringLayout.SOUTH, tempLabel);
		layout.putConstraint(SpringLayout.WEST,  photoLabel, 0,  SpringLayout.WEST,  rainLabel);
		layout.putConstraint(SpringLayout.NORTH, photoValue, -2, SpringLayout.NORTH, photoLabel);
		layout.putConstraint(SpringLayout.WEST,  photoValue, 10, SpringLayout.EAST,  transLabel);
		
		layout.putConstraint(SpringLayout.NORTH, transLabel, 20, SpringLayout.SOUTH, photoLabel);
		layout.putConstraint(SpringLayout.WEST,  transLabel, 0,  SpringLayout.WEST,  rainLabel);
		layout.putConstraint(SpringLayout.NORTH, transValue, -2, SpringLayout.NORTH, transLabel);
		layout.putConstraint(SpringLayout.WEST,  transValue, 10, SpringLayout.EAST,  transLabel);
		
		layout.putConstraint(SpringLayout.NORTH, cancelButton, 20, SpringLayout.SOUTH, transLabel);
		layout.putConstraint(SpringLayout.EAST,  cancelButton, 0, SpringLayout.EAST, transValue);
		layout.putConstraint(SpringLayout.NORTH, doButton, 0, SpringLayout.NORTH, cancelButton);
		layout.putConstraint(SpringLayout.EAST,  doButton, -10, SpringLayout.WEST, cancelButton);

		setLocationRelativeTo(parent);
		pack();
		
		show();
	}
	

	/** React to button presses */
	public void actionPerformed(ActionEvent e){
		if(e.getSource() == doButton){
			if(valuesOK()){ 
				sim.setRainfall(rainfall);
				sim.setTemperature(temperature);
				sim.setRadiation(radiation);
				sim.setTranspiration(transpiration);
				dispose();
				sim.setMode(Simulation.MODE_SIMULATE);
			}
		}
		if(e.getSource() == cancelButton){
			dispose();
			sim.setMode(Simulation.MODE_SIMULATE);
		}
	}
	

	/** Check if all input values are ok */
	public boolean valuesOK(){
		
		try{ rainfall = Double.parseDouble(rainValue.getText());
		}catch(NumberFormatException ex){
			Popup.inform("The rainfall value must be a number.");
			return false; }
		if(rainfall < 0 || rainfall > 500){
			Popup.inform("The rainfall value must be between 0 and 500");
			return false; }
		
		try{ temperature = Double.parseDouble(tempValue.getText());
		}catch(NumberFormatException ex){
			Popup.inform("The temperature value must be a number.");
			return false; }
		if(temperature < 0 || temperature > 50){
			Popup.inform("The temperature value must be between 0 and 50");
			return false; }
	
		try{ radiation = Double.parseDouble(photoValue.getText());
		}catch(NumberFormatException ex){
			Popup.inform("The value for photosynthetic radiation must be a number.");
			return false; }
		if(radiation < 0 || radiation > 60){
			Popup.inform("The value for photosynthetic radiation must be between 0 and 60");
			return false; }
		
		try{ transpiration = Double.parseDouble(transValue.getText());
		}catch(NumberFormatException ex){
			Popup.inform("The value for potential evapo-transpiration must be a number.");
			return false; }
		if(transpiration < 0 || transpiration > 250){
			Popup.inform("The value for potential evapo-transpiration must be between 0 and 250");
			return false; }
		
		return true;
	}

}
