/*
 * File:   SimulationState.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */


package sim;

import sim.pasture.*;
import sim.livestock.*;
import gui.*;
import utils.*;
import main.PastureSim;


import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import javax.swing.JOptionPane;


/** Represents the live state of the simulation. Deals with
  * the animation and timing of all simulation elements */
public final class SimulationState extends State{

	// Number of missed iterations allowed before slowing down
	public static final int ITERATION_SKIP_SLOWDOWN = 50;
	
	// Maximum allowed missed iteartions. If the number of missed itrations
	// surpasses this limit then the simulation rate will be non-linear
	// A beep will be sounded if this happens.
	public static final int ITERATION_SKIP_MAX = 100;
	
	// These variables keep track of simulation loop timing
	private transient boolean stopped;
	private transient boolean finished;
	private transient boolean endIteration;
	private transient int underStress;
	private transient long loopSpeed;
	
	
	/** Creates a fresh simulation object for when the application is first started */
	public SimulationState(Simulation sim, File imgFile){
		setName("Main Simulation Loop");
		this.sim = sim;
		fields = new LinkedList();
		gates = new LinkedList();
		water = new LinkedList();
		animals = new LinkedList();
		pState = new PastureState(sim, imgFile);
		Dimension d = sim.getPastureSize();
		pos = new Animal[d.width][d.height];
		waterPath = new PathFinderMap(); // Create a dummy water path
		ownedGates = new LinkedList();
	}
	
	/** Freezes this simulation to be used as a history item */
	public FrozenState freeze(){
		return new FrozenState(sim, fields, gates, water, animals, pState, totalTime); 
	}
	/** Creates a new simulation from a frozen one */
	public SimulationState(	Simulation sim,
				LinkedList fields,
				LinkedList gates,
				LinkedList water,
				LinkedList animals,
				PastureState pState,
				long totalTime){
		this.sim = sim;
		this.fields = fields;
		this.gates = gates;
		this.water = water;
		this.animals = animals;
		this.pState = pState;
		this.totalTime = totalTime;
		
		stopped = true;
		finished = false;
		endIteration = true;
		underStress = 0;
		loopSpeed = sim.getSpeed();
//System.out.println(""+this+": SETTING OWNED GATES");
	
		initialise(sim);
	}


/*	
	// Returns wheather we own this area
	public boolean ownsArea(Point p, int size){ return super.ownsArea(p, size); }
	// Returns any connected fields containing this point over an open gate
	public AnimalContainer findOpenNeighbours(Point p){ return super.findOpenNeighbours(p); }
*/
	//public void loadPastureStateImage(){ pState.loadImage(); }
	
/*
	public void setSimulation(Simulation sim){
		this.sim = sim;
		stopped = true;
		finished = false;
		endIteration = true;
		underStress = 0;
		loopSpeed = sim.getSpeed();
		for(Iterator i = fields.iterator(); i.hasNext();){ ((FieldBounds)i.next()).setSimulation(sim); }
		for(Iterator i = gates.iterator(); i.hasNext();){ ((FieldGate)i.next()).setSimulation(sim); }
		for(Iterator i = animals.iterator(); i.hasNext();){ ((LivestockGroup)i.next()).setSimulation(sim); }
	}
*/	
//	public void init(){
	
//	}

	
	//public long getTime(){ return timeMilli; }
	/*public String getReadableTime(){
		Calendar c = Calendar.getInstance();
		c.setTimeInMillis(timeMilli);
		return "" + c.get(Calendar.DAY_OF_MONTH) + "-" + c.get(Calendar.MONTH) + "-" + c.get(Calendar.YEAR)
		   + "  " + c.get(Calendar.HOUR) + ":" + c.get(Calendar.MINUTE) + ":" + c.get(Calendar.SECOND);
	}*/
	
/*	public void initPasture(){
	
	}
*/	

	public void play(){ stopped = false; }
	public void pause(){ stopped = true; pState.doFilter(); }
	public void kill(){ finished = true; }
	
	public void run(){
	
//System.out.println("SIM LOOP STATRED");
		finished = false;
		while(!finished){
			long timeBucket = 0, oldTime, time, timePassed, lostTime = 0, extraIterations = 0;
			double loopInterval = 1000;
			// Capture the current system time before the simulation starts
			time = System.currentTimeMillis();
			
			while(!stopped && !finished && sim.getSpeed() > 0){
				loopSpeed = sim.getSpeed();
				
				// Do the actual simulation
				simulationUpdate(extraIterations); 
				// Remove number of complete iterations from time bucket
				timeBucket -= (extraIterations+1)*(int)loopInterval;
				
				// Calculate the time which passed during the simulation update
				// Add on any time lost in the last iteration			
				oldTime = time;
				if(!sim.maxSpeed) time = System.currentTimeMillis();
//totalTime += (time - oldTime) * loopSpeed * sim.TIME_FACTOR;
				if(!sim.saving()){ sim.saveHistory(totalTime); }
		
				// Add passed time to the bucket
				if(!sim.maxSpeed) timeBucket += time - oldTime;
				else timeBucket = 0;				
				
				// Get the simulation loop interval
				if(loopSpeed > 0) loopInterval = 1000/loopSpeed;
				else break;
				
				// If we have time to spare then sleep
				if(timeBucket < 0){
					try{ sleep(-timeBucket); }
					catch(InterruptedException e){ e.printStackTrace(); }
					extraIterations = 0;
				}
				// If the bucket has more time in it then we will need to
				// complete more iterations next time
				else{
					// Get the number of extra iteartions which need to be
					// completed to empty the bucket
					if(loopInterval < 1) extraIterations = timeBucket;
					else extraIterations = (timeBucket/(int)loopInterval);

					// Prevent the bucket from overflowing
					if(extraIterations > ITERATION_SKIP_SLOWDOWN){
						Toolkit toolkit;
						toolkit = Toolkit.getDefaultToolkit();
						toolkit.beep(); 
						// If extraiterations is too large the bucket overflows
						// so just empty it to ease the load on the processor
						if(extraIterations > ITERATION_SKIP_MAX){
							//totalTime -= (extraIterations*loopInterval + lostTime);
							extraIterations = 0;
							timeBucket = 0;
							// We dropped iterations so sound the alarm with double beep
							toolkit.beep(); toolkit.beep();
						}
						sim.minusSpeed();
					}
				}
			}
			
			yield(); // Give other threads a chance
		}
	}
	

	public long getLoopSpeed(){ return loopSpeed; }
		
	// This is the the update which happens in each simulation iteration
	private void simulationUpdate(long n){

		if(sim.maxSpeed) n = 5; // If running at max speed, display 1 in 5
		
		// Do simulation stuff
		for(int i  = 0; i <= n; i++){
			if(!PastureSim.printPastureData && !PastureSim.printAnimalData) System.out.print("*");

			// Let the pastures re-generate
			pState.regenerate();

			// Let the animals feed
			for(Iterator j = animals.iterator(); j.hasNext();){
				((LivestockGroup)j.next()).animate();
			}
			
			// Incrament total time
			totalTime += 1000*sim.TIME_FACTOR;
		}

		// Draw the new state
		//sim.drawOnce(); 
		if(!main.PastureSim.textOnly) sim.reDraw(pState);
		
		// Garbage collect now
		//System.gc();
	}
	
		
/*	public void writeData(java.io.ObjectOutputStream out) throws IOException{
		//out.writeObject(gates);
		//out.writeObject(fields);
		//pState.setImageFile(sim.genCurrentFile());
		//out.writeObject(pState);
		//out.writeObject(animals);
		out.writeLong(totalTime);
	}
	
	public SimulationState(java.io.ObjectInputStream in, Simulation sim) throws IOException, ClassNotFoundException{
	
		this.sim = sim;
		Dimension d = sim.getPastureSize();
		this.pos = new Animal[d.width][d.height];
		gates = (LinkedList)in.readObject();
		fields = (LinkedList)in.readObject();
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			g.setSimulation(sim);
			for(Iterator j = fields.iterator(); i.hasNext();){
				FieldBounds f = (FieldBounds)j.next();
				f.setSimulation(sim);
				f.notifyGateArrival(g);	
			}	
		}
		//pState = (PastureState)in.readObject();
		animals = (LinkedList)in.readObject();
		for(Iterator i = animals.iterator(); i.hasNext();){
			LivestockGroup l = (LivestockGroup)i.next();
			//l.setData(sim, pos, pState);
		}
		totalTime = in.readLong();
		
		stopped = true;
		finished = false;
		endIteration = true;
		underStress = 0;
		loopSpeed = sim.getSpeed();
	}
		
*/
/*
	private void writeObject(java.io.ObjectOutputStream out) throws IOException{
		out.writeObject(fields);
		out.writeObject(gates);
		out.writeObject(animals);
//System.out.println("ATTEMPTIMG TO WRITE PSTATE OBJECT)");
//pState.setImageFile(sim.genCurrentFile());
//pState.writeObject(out);
		out.writeObject(pState);
//		out.writeObject(pos);
		out.writeLong(totalTime);
	}
	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException{
		fields = (LinkedList)in.readObject();
		gates = (LinkedList)in.readObject();
		animals = (LinkedList)in.readObject();
		pState = (PastureState)in.readObject();
//		pos = (Animal[][])in.readObject();
		totalTime = in.readLong();
	}
*/	

	
	public String toString(){ return "sim.SimulationState ["+pState+"]"; }
}

