/*
 * File:   Animal.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */

 
 
package sim.livestock;

import sim.*;
import sim.pasture.*;
import utils.*;


import java.io.*;
import java.awt.*;
import java.awt.geom.*;
import java.util.Vector;


/** All animals in the simulations are subclasses of this */
public abstract class Animal extends Point2D.Double implements Serializable{

	// Animal types
	public static final int COW = 0;
	public static final int SHEEP = 1;
	public static final int HORSE = 2;
	
	public static final Color COLOR_COW = new Color(255,200,0);
	public static final Color COLOR_SHEEP = Color.WHITE;
	public static final Color COLOR_PIG = Color.BLACK;
	public static final Color COLOR_ACTIVE = Color.LIGHT_GRAY;

	
	// Physological states
	public static final int STATE_EARLY     = 0; // Early growth
	public static final int STATE_LATE      = 1; // Late growth
	public static final int STATE_PREGNANT  = 2; // Pregnant
	public static final int STATE_LACTATING = 3; // Lactating
	public static final int STATE_ADULT     = 4; // Adult
	
	
	// Behaviour bits for the desicion variable
	public static final int DO_NOTHING  = 0;
	public static final int EAT         = 1;
	public static final int FORAGE      = 2;
	public static final int DRINK       = 3;
	public static final int HERD        = 4;
	public static final int MEMORY      = 5;
	public static final int LMEMORY     = 6;
	
	
	protected transient Simulation sim;         // A reference to the simulation
	protected transient LivestockGroup herd;    // A reference to the herd
	protected transient FieldGeometry geom;     // The field geometry
	protected transient PastureState grass;     // A reference to the pasture
	protected transient AnimalContainer bounds; // The animal's containing field
	protected transient Animal[][] pos;         // The position array for all animals
	protected           boolean sex;            // The animal's sex
	protected           int age;                // The animal's age
	protected           float energy;           // The level of energy
	protected           float hydration;        // The level of hydration in litres
	protected           int minutesPassed;      // The number of minutes passed within a year
	protected           Vector2D direction;     // The animals orientation
	protected           boolean alive;          // Wheather the animal is alive
	
	
	/** Constructor for a new animal */
	public Animal(Simulation sim, double px, double py, LivestockGroup herd){
		this.sim = sim;
		this.herd = herd;
		x = px;
		y = py;
		minutesPassed = 0;
		geom = sim.getFieldGeometry();
		bounds = geom.findContainingField((Point2D.Double)this);
		grass = sim.getPastureState();
		pos = sim.getPos();
		pos[(int)x][(int)y] = this;
		direction = new Vector2D();
	}
	
	
	/** Initialise global references after a load */
	public void initialise(Simulation sim, Animal[][] pos, PastureState pState, LivestockGroup herd){
		this.pos = pos;
		initialise(sim, pState, herd);
		if(sim.withinPastureBounds((int)x,(int)y)) pos[(int)x][(int)y] = this;
	}
	
	/** Initialise global references after a load */
	public void initialise(Simulation sim, PastureState pState, LivestockGroup herd){
		this.sim = sim;
		this.geom = sim.getFieldGeometry();
		this.bounds = geom.findContainingField((Point2D.Double)this);
		this.grass = pState;
		this.herd = herd;
	}

	
	
	public float energy(){ return energy; }	
	public float hydration(){ return hydration; }	
/*	
	public abstract Object clone(PastureState grass, Animal[][]pos);
	public Animal(Simulation sim, FieldGeometry geom,
	                      PastureState grass, boolean sex, int age,
			      int energy, int hydration, boolean alive,
			      int minutesPassed, Animal[][] pos, Vector2D direction, LivestockGroup herd){
		
		this.sim = sim;
		this.geom = geom;
		this.grass = grass;
		this.sex = sex;
		this.age = age;
		this.energy = energy;
		this.hydration = hydration;
		this.alive = alive;
		this.minutesPassed = minutesPassed;
		this.bounds = null; // Rememer to call geom.findContainingField
		this.pos = pos;
		this.pos[(int)x][(int)y] = this;
		this.direction = direction;
		this.herd = herd;
	}
		       
	public void setSimulation(Simulation sim){
		this.sim = sim;
		pos = sim.getPos();
		pos[(int)x][(int)y] = this;
	}

*/	

	public void notifyFieldRemoval(AnimalContainer f){ if(f == bounds) bounds = sim.state; }
	
	protected void animate(){
		minutesPassed++;
		if(minutesPassed == 525960){
			minutesPassed = 0;
			age++;
		}
	}
	
    public boolean isAlive(){ return alive; }

	protected abstract void roam();
	
	/** Move the animal if it is anywere that it shouldn't be */
	protected void roam(Point2D.Double np){
		
		double oldX = x, oldY = y;
		preventStray(np);
		disperse();
				
		// Update pos array with new values
		if(withinArea(oldX, oldY))
			pos[(int)oldX][(int)oldY] = null;
		if(withinArea(x, y))
			pos[(int)x][(int)y] = this;
	}
	
	/** Stops the animal from leaving a field, or entering a water area. */
	private void preventStray(Point2D.Double np){
	
		// Check if the animal is going into water
		WaterGeometry wgeom = sim.getWaterGeometry();
		if(wgeom.inWater(np)) return;
			
		// Check if the animal is not leaving its field
		if(bounds.holds(np)){ x = np.x; y = np.y; }
		else{
			AnimalContainer desiredField = geom.findContainingField(np);
			if(bounds.connected(desiredField, new Line2D.Double((Point2D.Double)this, np))){
				x = np.x; y = np.y;
				bounds = desiredField;
			}
		}
	}
	
	/** Disperse the animals if too close */
	public void disperse(){
	
		// Using ceiling to ensure that animals will definetly move initially
		int scanArea = (int)((sim.ANIMAL_SIZE+1.0)/2.0);
		int x = (int)this.x;
		int y = (int)this.y;
		
		if(withinArea(x, y)
		   && pos[x][y] != this
		   && pos[x][y] != null){
			separate(pos[x][y]);
			return;
		}
			
		//System.out.print("SCAN AREA = " + scanArea + "ANIMAL_SIZE = " + sim.ANIMAL_SIZE);

		// Scans for the closest animal which is closer than scanArea
		// The animals may appear too scattered if using a large pasture
		// area whilst zoomed in. This is because of the limitation that
		// each animal must be in its own pixel, to extend the capability
		// we would use a linked list array insted of an Animal array for pos
		// but that would probably be too in efficent.
		for(int i = 1; i <= scanArea; i++){
			
			Vector cand = new Vector();
			if(withinArea(x+i,y) && pos[x+i][y] != null && pos[x+i][y] != this) cand.add(pos[x+i][y]);
			if(withinArea(x-i,y) && pos[x-i][y] != null && pos[x-i][y] != this) cand.add(pos[x-i][y]);
			if(withinArea(x,y+i) && pos[x][y+i] != null && pos[x][y+i] != this) cand.add(pos[x][y+i]);
			if(withinArea(x,y-i) && pos[x][y-i] != null && pos[x][y-i] != this) cand.add(pos[x][y-i]);
			if(separationDone(cand)) return;
			
			for(int j = 1; j < (int)Math.floor(i/2); j++){
				
				if(distance(new Point2D.Double(x+j, y+i)) < scanArea){
					cand = new Vector();
					if(withinArea(x+j,y+i) && pos[x+j][y+i] != null && pos[x+j][y+i] != this) cand.add(pos[x+j][y+i]);
					if(withinArea(x+j,y-i) && pos[x+j][y-i] != null && pos[x+j][y-i] != this) cand.add(pos[x+j][y-i]);
					if(withinArea(x-j,y+i) && pos[x-j][y+i] != null && pos[x-j][y+i] != this) cand.add(pos[x-j][y+i]);
					if(withinArea(x-j,y-i) && pos[x-j][y-i] != null && pos[x-j][y-i] != this) cand.add(pos[x-j][y-i]);
					if(withinArea(x+i,y+j) && pos[x+i][y+j] != null && pos[x+i][y+j] != this) cand.add(pos[x+i][y+j]);
					if(withinArea(x+i,y-j) && pos[x+i][y-j] != null && pos[x+i][y-j] != this) cand.add(pos[x+i][y-j]);
					if(withinArea(x-i,y+j) && pos[x-i][y+j] != null && pos[x-i][y+j] != this) cand.add(pos[x-i][y+j]);
					if(withinArea(x-i,y-j) && pos[x-i][y-j] != null && pos[x-i][y-j] != this) cand.add(pos[x-i][y-j]);
					if(separationDone(cand)) return;
				}
			}
			
			if(distance(new Point2D.Double(x+i, y+i)) < scanArea){
				cand = new Vector();
				if(withinArea(x+i,y+i) && pos[x+i][y+i] != null && pos[x+i][y+i] != this) cand.add(pos[x+i][y+i]);
				if(withinArea(x+i,y-i) && pos[x+i][y-i] != null && pos[x+i][y-i] != this) cand.add(pos[x+i][y-i]);
				if(withinArea(x-i,y+i) && pos[x-i][y+i] != null && pos[x-i][y+i] != this) cand.add(pos[x-i][y+i]);
				if(withinArea(x-i,y-i) && pos[x-i][y-i] != null && pos[x-i][y-i] != this) cand.add(pos[x-i][y-i]);
				if(separationDone(cand)) return;
			}
		}
	}
	
	// Finds the closest animal in the vector and seperates it
	public boolean separationDone(Vector v){
		if(v.size() == 0) return false;
		Animal choice = (Animal)v.get((int)(Math.random()*v.size()));
		separate(choice);
		return true;
	}
	
	
	public boolean withinArea(double x, double y){
		return x >= 0 && x < pos.length 
		       && y >= 0 && y < pos[0].length;
	}
	
	// Remove this animal from the pos array once it has died
	public void removePos(){
		if(sim.withinPastureBounds((int)x,(int)y))
			pos[(int)x][(int)y] = null;
	}
	
	public final void notifyNewField(AnimalContainer f){
		if(f.holds((Point2D.Double)this)) bounds = f;
	}
	
	/** Draw hunger data */
	public abstract void drawHunger(Graphics g);
	
	/** Draw thirst data */
	public abstract void drawThirst(Graphics g);
	
	/** Draws a live animal */
	public void draw(Graphics g){
		sim.fillScaledOval(g, x-sim.ANIMAL_SIZE/2, y-sim.ANIMAL_SIZE/2, sim.ANIMAL_SIZE, sim.ANIMAL_SIZE);
		Point2D.Double head = direction.project(x, y, sim.ANIMAL_SIZE/2.0);
		sim.fillScaledOval(g, head.x-sim.ANIMAL_SIZE/4.0, head.y-sim.ANIMAL_SIZE/4.0, sim.ANIMAL_SIZE/2.0, sim.ANIMAL_SIZE/2.0);
	}

	/** Drawas a dead animal */
	public void drawDead(Graphics g){
		sim.fillScaledOval(g, x-sim.ANIMAL_SIZE/4, y-sim.ANIMAL_SIZE/4, sim.ANIMAL_SIZE/2, sim.ANIMAL_SIZE/2);
		if(hydration <= 0 && energy <= 0) g.setColor(Color.BLACK);
		else if(energy <= 0) g.setColor(Color.RED);
		else if(hydration <= 0) g.setColor(Color.BLUE);
		else g.setColor(Color.GREEN);
		sim.drawScaledLine(g, x-sim.ANIMAL_SIZE/2, y-sim.ANIMAL_SIZE/2, x+sim.ANIMAL_SIZE/2, y+sim.ANIMAL_SIZE/2);
		sim.drawScaledLine(g, x-sim.ANIMAL_SIZE/2, y+sim.ANIMAL_SIZE/2, x+sim.ANIMAL_SIZE/2, y-sim.ANIMAL_SIZE/2);
	}
	
	public void place(Point2D.Double p){
		x = p.x; y = p.y;
		bounds = geom.findContainingField((Point2D.Double)this);
	}
	
	public final void separate(Animal a){
		Vector2D v;
		v = new Vector2D(a.x, a.y, x, y);
		while(v.isZeroVector()){ v = new Vector2D(); } // Use random direction
		// Move away from the closest animal
		Point2D.Double next = null;
		//if((int)distance(a) < sim.ANIMAL_SIZE) next = v.project(x, y, sim.ANIMAL_SIZE/(int)distance(a)); //Math.sqrt(sim.ANIMAL_SIZE/2  sim.ANIMAL_SIZE/2
		//else 
		//System.out.println(ANIMAL_SIZE);
		next = v.project(x, y, sim.ANIMAL_SIZE/4.0);
		direction = v;
		preventStray(next);
	}
	

	private void writeObject(java.io.ObjectOutputStream out) throws IOException{
		out.writeDouble(x);
		out.writeDouble(y);
		out.writeBoolean(sex);
		out.writeBoolean(alive);
		out.writeInt(age);
		out.writeFloat(energy);
		out.writeFloat(hydration);
		out.writeInt(minutesPassed);
		out.writeObject(direction);
		//out.writeObject(herdCentre);
		//out.writeInt(herdSize);
	}
	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException{
		x             = in.readDouble();
		y             = in.readDouble();
		sex           = in.readBoolean();
		alive         = in.readBoolean();
		age           = in.readInt();
		energy        = in.readFloat();
		hydration     = in.readFloat();
		minutesPassed = in.readInt();
		direction     = (Vector2D)in.readObject();
		//herdCentre = (Point)in.readObject();
		//herdSize = in.readInt();
	}

	
}
