/*
 * File:   FieldBounds.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */

 
package sim.pasture;

import gui.*;
import sim.pasture.*;
import sim.livestock.*;
import sim.*;
import utils.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.io.*;
import java.util.*;
import java.awt.geom.*;


/** This represents the pasture's fields as a polygon */
public class FieldBounds extends Polygon implements FieldItem, Serializable, AnimalContainer{

	private transient LinkedList gates;        // A reference to all gates
	private transient PathFinderMap waterPath; // The paths to the water
	private transient Simulation sim;          // A reference to the simulation object

	// Used to give the fields an id, for debug purposes
	public static int field_count;
	private int id;
	
	
	/** Constructor for a new field */
	public FieldBounds(Polygon p, Simulation sim){
		npoints = p.npoints;
		xpoints = p.xpoints;
		ypoints = p.ypoints;
		this.sim = sim;
		gates = new LinkedList();
		id = field_count++;
	}
	
	/** Called to re-set global references after a load */
	public void initialise(Simulation sim, LinkedList gates){
		this.sim = sim;
		this.gates = new LinkedList();
		// Store the gates which we own
		for(Iterator i = gates.iterator(); i.hasNext(); ){
			FieldGate g = (FieldGate)i.next();
			if(this.ownsGate(g)) this.gates.add(g);
		}
	}
	
	
	
	
	// ---------------[ Stuff for drawing fields ]---------------
	
	public void draw(Graphics g){
		g.setColor(Color.YELLOW);
		drawShape(g);
	}
	
	public void drawActive(Graphics g){
		g.setColor(Color.WHITE);
		drawShape(g);
	}
	
	public void drawShape(Graphics g){
		int i;
		for(int ni = 0; ni < npoints; ni++){
			if(ni == 0) i = npoints-1;
			else i = ni-1;
			LinkedList pg = getPossessedGates(xpoints[i], ypoints[i], xpoints[ni], ypoints[ni]);
			drawFence(g, xpoints[i], ypoints[i], xpoints[ni], ypoints[ni], pg);
		}
	}
	
	public void drawFence(Graphics g, double x, double y, int xlast, int ylast, LinkedList pg){
		Line2D.Double gate = null;
		if(pg != null){
			while(!pg.isEmpty()){
				gate = popClosestGate(pg, x, y);
				sim.drawScaledLine(g, x, y, gate.x1, gate.y1);
				x = gate.x2; y = gate.y2;
			}
		}
		sim.drawScaledLine(g, x, y, xlast, ylast);
	}

	public Line2D.Double popClosestGate(LinkedList gates, double x, double y){
		double d = Double.MAX_VALUE; // Smallest dist so far
		Point2D.Double p = new Point2D.Double(x,y);
		Line2D.Double g = null; // Closest gate so far
		FieldGate bg = null; // Best gate so far
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate fg = (FieldGate)i.next();
			Line2D.Double seg = fg.getSegment();
			Point2D.Double p1 = new Point2D.Double(seg.x1, seg.y1);
			Point2D.Double p2 = new Point2D.Double(seg.x2, seg.y2);
			double d_ = p.distance(p1);
			if(d_ < d){ d = d_; g = new Line2D.Double(p1, p2); bg = fg; }
			d_ = p.distance(p2);
			if(d_ < d){ d = d_; g = new Line2D.Double(p2, p1); bg = fg; }
		}
		gates.remove(bg);
		return g;
	}
	
	// ---------[ Stuff for calculating gates and bounds ]-------
	
	
	public LinkedList getPossessedGates(int x1, int y1, int x2, int y2){
		if(gates == null) return null;
		LinkedList pg = new LinkedList();
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate fg = (FieldGate)i.next();
			if(fg.onLine(x1,y1,x2,y2)){ pg.add(fg); }
		}
		return pg;
	}
	
	
	// This implements animal Container	
	public boolean ownsArea(Point2D.Double p, int size){ return contains(p.x, p.y, size, size); }
	
	// Returns the animal container containing the point p which is over an open owned gate
	public AnimalContainer findOpenNeighbours(Point2D.Double p){
		for(Iterator i = gates.iterator(); i.hasNext(); ){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen() && g.near(p)) return g.neighbour((AnimalContainer)this);
		}
		return null;
	}
	
	// Returns whether this field fully contains p
	public boolean holds(Point2D.Double p){ return super.contains(p.x,p.y); }
	
	// Regurns wether the fild holds the pixel p
	public boolean fullyHolds(Point2D.Double p){
		return super.contains(new Rectangle((int)p.x, (int)p.y, 1, 1)); 
	}

	/** Returns true if the animal containers are connected by an open gate accross the line l */
	public boolean connected(AnimalContainer ac, Line2D.Double l){
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen() && g.neighbour(this) == ac && g.lineFits(l)) return true;
		}
		return false;
	}
	

	/** Adds the gate if we own it */
	public void notifyGateArrival(FieldGate g){ if(this.ownsGate(g)) gates.add(g); }
	/** Removes the gate if we own it */
	public void notifyGateRemoval(FieldGate g){ gates.remove(g); }
	
	/** Adds the gates which we own */
	public void addExistingGates(LinkedList g){
		FieldGate fg = null;
		for(Iterator i = g.iterator(); i.hasNext();){
			fg = (FieldGate)i.next();
			if(this.ownsGate(fg)) gates.add(fg);
		}
	}

				
	public boolean overlapPoints(Polygon p){
		for(int i = 0; i < p.npoints; i++){
			if(this.contains(new Point(p.xpoints[i], p.ypoints[i]))){ 
//System.out.println(this);
//System.out.println("("+p.xpoints[i]+","+p.ypoints[i]+")");
			return true; }
		}
		return false;
	}
	
	
	public boolean potentialOverlap(Polygon p){
		if(overlapPoints(p)) return true;
		for(int i = 0; i < npoints-1; i++){
			for(int j = 0; j < p.npoints-1; j++){
				
				if(XLine2D.linesIntersect(xpoints[i], ypoints[i], xpoints[i+1], ypoints[i+1],
				   p.xpoints[j], p.ypoints[j], p.xpoints[j+1], p.ypoints[j+1])){
//System.out.println("potential overlap ("+xpoints[i]+","+ypoints[i]+")("+xpoints[i+1]+","+ypoints[i+1]+") - ("+p.xpoints[j]+","+p.ypoints[j]+")("+p.xpoints[j+1]+","+p.ypoints[j+1]+")");
				   
				   return true; }
			}
		}
		return false;
	}
	
	public boolean overlap(Polygon p){
		//if(overlapPoints(p)) return true;
		for(int i = 0; i < npoints-1; i++){
			for(int j = 0; j < p.npoints-1; j++){
				if(XLine2D.linesIntersect(xpoints[i], ypoints[i], xpoints[i+1], ypoints[i+1],
				   p.xpoints[j], p.ypoints[j], p.xpoints[j+1], p.ypoints[j+1])){
//System.out.println("overlap ("+xpoints[i]+","+ypoints[i]+")("+xpoints[i+1]+","+ypoints[i+1]+") - ("+p.xpoints[j]+","+p.ypoints[j]+")("+p.xpoints[j+1]+","+p.ypoints[j+1]+")");
				   return true; }
			}
		}
		return false;
	}
	
	
	public Point getCloseVertex(Point2D.Double pt){
		Point p = null; // Closest point
		double d = sim.SNAP_RADIUS; // Smallest distance
		for(int i = 0; i < npoints; i++){
			double d_ = Point2D.distance(xpoints[i], ypoints[i], pt.x, pt.y); 
			if(d_ < d){
				d = d_;
				p = new Point(xpoints[i], ypoints[i]);
			}
		}
		return p;
	}

	public double getClosestFence(double x, double y, Point p1, Point p2){
		double d = Double.MAX_VALUE; // Smallest distance so far
		int i;
		for(int ni = 0; ni < npoints; ni++){
			if(ni == 0) i = npoints-1;
			else i = ni-1;
			double d_ = Line2D.ptLineDist(xpoints[i], ypoints[i], xpoints[ni], ypoints[ni], x, y); 
			if(d_ < d){
				d = d_;
				if(Line2D.relativeCCW(xpoints[i], ypoints[i], xpoints[ni], ypoints[ni], x, y) < 0){
					p1.x = xpoints[i]; p1.y = ypoints[i]; p2.x = xpoints[ni]; p2.y = ypoints[ni];
				}else{ p1.x = xpoints[ni]; p1.y = ypoints[ni]; p2.x = xpoints[i]; p2.y = ypoints[i]; }
			}
		}
		return d;
	}
		
	public boolean contains(Point2D.Double p){
		for(int i = 0; i < npoints; i++){
			if((int)p.x == xpoints[i] && (int)p.y == ypoints[i]) return false;
		}	
		return super.contains(p.x, p.y);
	}

	public boolean contains(Point p){
		for(int i = 0; i < npoints; i++){
			if(p.x == xpoints[i] && p.y == ypoints[i]) return false;
		}	
		return super.contains(p.x, p.y);
	}
	
	public boolean ownsGate(FieldGate g){
		int i;
		for(int ni = 0; ni < npoints; ni++){
			if(ni == 0) i = npoints-1;
			else i = ni-1;
			if(g.onLine(xpoints[i], ypoints[i], xpoints[ni], ypoints[ni])) return true;
		}
		return false;
	}
	
	// Returns true if this field is connected to f via an open gate near point a
	public boolean connected(FieldBounds f, Point a){
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen() && a.distance(g) < sim.GATE_WIDTH/2 && (f == null || f.ownsGate(g))){
				return true;
			}
		}
		return false;
	}
	
	
	public FieldGate getOpenConnectingGate(FieldBounds f, Point a){
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen() && a.distance(g) < sim.GATE_WIDTH/2 && (f == null || f.ownsGate(g)))
				return g;
		}
		return null;
	}

	
	// Returns true if this field is connected to f via a gate near point a
	public boolean connectedGate(FieldBounds f, Point a){
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(a.distance(g) < sim.GATE_WIDTH/2 && (f == null || f.ownsGate(g))){
				return true;
			}
		}
		return false;
	}
	
	
	/** Returns weather the point p is on any of the field's gates */
	public boolean onGate(Point2D.Double p){
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(p.distance(g) < sim.GATE_WIDTH*0.5){
				return true;
			}
		}
		return false;
	}

	
	/** Test wheather the line (x1,y1)-(x2,y2) intersects any of the edges */
	public boolean intersects(Point p1, Point p2){
		int i;
		for(int ni = 0; ni < npoints; ni++){
			if(ni == 0) i = npoints-1;
			else i = ni-1;
			if(Line2D.linesIntersect(p1.x, p1.y, p2.x, p2.y, xpoints[i], ypoints[i], xpoints[ni], ypoints[ni])) return true;
		}
		return false;
	}
	
	
	// ------------[ stuff for calculating water ]----------
	
	/** Returns all water accessible from this field */
	public LinkedList findAvailableWater(LinkedList allWater){
		LinkedList availWater = new LinkedList();
		for(Iterator i = allWater.iterator(); i.hasNext();){
			FieldWater w = (FieldWater)i.next();
			if(overlap(w)) availWater.add(w);
		}
		return availWater;
	}
	
	
	public String toString(){ 
		String s = "FieldBounds["+id+"]";
		for(int i = 0; i < npoints; i++)
			s = s+"("+xpoints[i]+","+ypoints[i]+")";
		return s;
	}
	
	// ------------[ Stuff for finding the best path to a resource ] -----------------
	

	// Finds the gate leading to the shortest path to the target
	public FieldGate findGate(Point2D.Double tgt){
		AnimalContainer tgtBounds = sim.getFieldGeometry().findContainingField(tgt);
		if(tgtBounds == null) return null;
		
		TreeSet cand = tgtBounds.fringeEdges(tgt);
		LinkedList visited = new LinkedList();
		
		while(cand.size() > 0){
			DijkstraEdge first = (DijkstraEdge)cand.first(); // Get the closest candidate gate
			if(first.next == this) return first.gate; // If its connecting field is this one then we are done
			else{ 
				cand.remove(first); // If not, then remove it
				visited.add(first.gate); // add it to the visited list
				cand.addAll(first.next.fringeEdges(first, visited)); // Then add all its field's candidates
			}
		}
		return null;
	}
	
	
	// Returns all dijkstra edges (gates) which are reachable
	// from the point p through this field, ordered by distance
	public TreeSet fringeEdges(Point2D.Double p){
		TreeSet edges = new TreeSet();
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen()){
				edges.add(new DijkstraEdge(g, g.distance(p), this));
			}
		}
		return edges;
	}
	
	
	// Returns all dijkstra edges (gates) which are reachable
	// from the gate e through this field, ordered by distance
	public TreeSet fringeEdges(DijkstraEdge e, LinkedList visited){
		TreeSet edges = new TreeSet();
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen() && !visited.contains(g)){
				edges.add(new DijkstraEdge(g, g.distance(e.gate)+e.weight, this));
			}
		}
		return edges;
	}
	
}
