/*
 * File:   FieldWater.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */

 
package sim.pasture;

import gui.*;
import sim.pasture.*;
import sim.livestock.*;
import sim.*;
import utils.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.io.*;
import java.util.*;
import java.awt.geom.*;


/** This represents the pasture's water soruces */
public class FieldWater extends Polygon implements FieldItem, Serializable{

	private static final Color WATER_COLOR = new Color(0, 127, 255);
	
	private transient Simulation sim;        // A reference to the simulation
	public transient Point2D.Double centre; // The centre of mass
	
	public FieldWater(Polygon p, Simulation sim){
		npoints = p.npoints;
		xpoints = p.xpoints;
		ypoints = p.ypoints;
		Rectangle bounds = p.getBounds();
		// Centre of mass will be the centre of the containing rectangle
		centre = new Point2D.Double( bounds.x+bounds.width/2.0, bounds.y+bounds.height/2.0 );
		// If that isnt actually in the polygon then fall back to the first vertex
		if(!contains(centre)){ centre = centreScan(centre, bounds.width, bounds.height); }
		this.sim = sim;
	}
	
	/** Initialise the water after a load */
	public void initialise(Simulation sim){
		this.sim = sim;
		
		Rectangle bounds = getBounds();
		// Centre of mass will be the centre of the containing rectangle
		centre = new Point2D.Double( bounds.x+bounds.width/2.0, bounds.y+bounds.height/2.0 );
	}

	
	// Scans outwards until it hits this polygon and returns the first collision
	public Point2D.Double centreScan(Point2D.Double p, int width, int height){
	
		int scanArea;
		int x = (int)(p.x+0.5);
		int y = (int)(p.y+0.5);
		
		if(width < height) scanArea = (int)width/2;
		else scanArea = (int)height/2;	
	
		for(int i = 1; i <= scanArea; i++){
			
			if(contains(x+i,y)) return new Point2D.Double(x+i,y);
			if(contains(x-i,y)) return new Point2D.Double(x-i,y);
			if(contains(x,y+i)) return new Point2D.Double(x,y+i);
			if(contains(x,y-i)) return new Point2D.Double(x,y-i);
			
			for(int j = 1; j < (int)Math.floor(i/2); j++){
				
				if(contains(x+j,y+i)) return new Point2D.Double(x+j,y+i);
				if(contains(x+j,y-i)) return new Point2D.Double(x+j,y-i);
				if(contains(x-j,y+i)) return new Point2D.Double(x-j,y+i);
				if(contains(x-j,y-i)) return new Point2D.Double(x-j,y-i);
				if(contains(x+i,y+j)) return new Point2D.Double(x+i,y+j);
				if(contains(x+i,y-j)) return new Point2D.Double(x+i,y-j);
				if(contains(x-i,y+j)) return new Point2D.Double(x-i,y+j);
				if(contains(x-i,y-j)) return new Point2D.Double(x-i,y-j);
			}
			
			if(contains(x+i,y+i)) return new Point2D.Double(x+i,y+i);
			if(contains(x+i,y-i)) return new Point2D.Double(x+i,y-i);
			if(contains(x-i,y+i)) return new Point2D.Double(x-i,y+i);
			if(contains(x-i,y-i)) return new Point2D.Double(x-i,y-i);
		}
		
		// This should never happen, but if scan fails default to first vertex.
		return new Point2D.Double(xpoints[0], ypoints[0]);
	}
	
	
	public void reInitialise(Simulation sim){
		this.sim = sim;
	}
	
	public void draw(Graphics g){
		g.setColor(WATER_COLOR);
		sim.fillScaledPolygon(g, xpoints, ypoints, npoints);
	}
	
	public void drawActive(Graphics g){
		g.setColor(Color.WHITE);
		drawShape(g);
	}
	
	public void drawShape(Graphics g){
		int i;
		for(int ni = 0; ni < npoints; ni++){
			if(ni == 0) i = npoints-1;
			else i = ni-1;
			sim.drawScaledLine(g, xpoints[i], ypoints[i], xpoints[ni], ypoints[ni]);
		}
	}
	
	
	
	public boolean overlapPoints(Polygon p){
		for(int i = 0; i < p.npoints; i++){
			if(contains(p.xpoints[i], p.ypoints[i])){ return true; }
		}
		return false;
	}
	
	
	public boolean potentialOverlap(Polygon p){
		if(overlapPoints(p)) return true;
		for(int i = 0; i < npoints-1; i++){
			for(int j = 0; j < p.npoints-1; j++){
				if(XLine2D.linesIntersect(xpoints[i], ypoints[i], xpoints[i+1], ypoints[i+1],
				   p.xpoints[j], p.ypoints[j], p.xpoints[j+1], p.ypoints[j+1])){ return true; }
			}
		}
		return false;
	}
	
	public boolean overlap(Polygon p){
		//if(overlapPoints(p)) return true;
		for(int i = 0; i < npoints-1; i++){
			for(int j = 0; j < p.npoints-1; j++){
				if(XLine2D.linesIntersect(xpoints[i], ypoints[i], xpoints[i+1], ypoints[i+1],
				   p.xpoints[j], p.ypoints[j], p.xpoints[j+1], p.ypoints[j+1])){
				   	//pa.d1A = new Point(xpoints[i], ypoints[i]);
					//pa.d1B = new Point(xpoints[i+1], ypoints[i+1]);
					//pa.d2A = new Point(p.xpoints[j], p.ypoints[j]);
					//pa.d2B = new Point(p.xpoints[j+1], p.ypoints[j+1]);
					return true; }
			}
		}
		return false;
	}
	
	/** Returns the polygon vertex which is closest to (x,y) and within snap range */
	public Point getCloseVertex(Point2D.Double pt){
		Point p = null; // Closest point
		double d = sim.SNAP_RADIUS; // Smallest distance
		for(int i = 0; i < npoints; i++){
			double d_ = Point2D.distance(xpoints[i], ypoints[i], pt.x, pt.y); 
			if(d_ < d){
				d = d_;
				p = new Point(xpoints[i], ypoints[i]);
			}
		}
		return p;
	}

	/** Returns the distance to the closest vertex to (x,y) and stores the result in p  */
	public double getClosestVertex(int x, int y, Point p){
		Point p_ = null; // Closest point
		double d = Double.MAX_VALUE; // Smallest distance
		for(int i = 0; i < npoints; i++){
			double d_ = Point2D.distance(xpoints[i], ypoints[i], x, y); 
			if(d_ < d){
				d = d_;
				p_ = new Point(xpoints[i], ypoints[i]);
			}
		}
		// Store result in p
		if(p_ != null){ p.x = p_.x; p.y = p_.y; }
		return d;
	}
	
	/** Returns the distance to the closest water border, storing the closest line in p1 and p2 */
	public double getClosestEdge(int x, int y, Point p1, Point p2){
		double d = Double.MAX_VALUE; // Smallest distance so far
		int i;
		for(int ni = 0; ni < npoints; ni++){
			if(ni == 0) i = npoints-1;
			else i = ni-1;
			double d_ = Line2D.ptLineDist(xpoints[i], ypoints[i], xpoints[ni], ypoints[ni], x, y); 
			if(d_ < d){
				d = d_;
				if(Line2D.relativeCCW(xpoints[i], ypoints[i], xpoints[ni], ypoints[ni], x, y) < 0){
					p1.x = xpoints[i]; p1.y = ypoints[i]; p2.x = xpoints[ni]; p2.y = ypoints[ni];
				}else{ p1.x = xpoints[ni]; p1.y = ypoints[ni]; p2.x = xpoints[i]; p2.y = ypoints[i]; }
			}
		}
		return d;
	}
	
	/** Returns the distance to the closest water border within the fieldbounds f, storing the closest line in p1 and p2 */
	public double getClosestEdge(int x, int y, Point p1, Point p2, FieldBounds f){
		double d = Double.MAX_VALUE; // Smallest distance so far
		int i;
		for(int ni = 0; ni < npoints; ni++){
			if(ni == 0) i = npoints-1;
			else i = ni-1;
			if(!f.intersects(x,y, xpoints[i], ypoints[i]) || !f.intersects(x, y, xpoints[ni], ypoints[ni])){
				double d_ = Line2D.ptLineDist(xpoints[i], ypoints[i], xpoints[ni], ypoints[ni], x, y); 
				if(d_ < d){
					d = d_;
					if(Line2D.relativeCCW(xpoints[i], ypoints[i], xpoints[ni], ypoints[ni], x, y) < 0){
						p1.x = xpoints[i]; p1.y = ypoints[i]; p2.x = xpoints[ni]; p2.y = ypoints[ni];
					}else{ p1.x = xpoints[ni]; p1.y = ypoints[ni]; p2.x = xpoints[i]; p2.y = ypoints[i]; }
				}
			}
		}
		return d;
	}
	
	// Checks wheather the water area will contain the point
	// once projected by vector v
	public boolean contains(Point2D.Double p, Vector2D v){
		return contains(v.project(p,1.5)); //Magnitude is root 2, ensures that we project by at least 1 full pixel
	}
	public boolean contains(Point2D.Double p){ return contains(p.x, p.y); }
	public boolean contains(int x, int y){
		for(int i = 0; i < npoints; i++){
			if(x == xpoints[i] && y == ypoints[i]) return false;
		}	
		return super.contains(x,y);
	}

}
