/*
 * File:   PastureState.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */


package sim.pasture;

import sim.*;
import sim.model.*;
import utils.GraphData;
import main.PastureSim;
import gui.Popup;

import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;
import javax.imageio.*;
import javax.imageio.stream.*;
import java.io.*;
import java.util.*;

public class PastureState implements Cloneable, Serializable{ //{

	public static final int VISIT_STAND = 0;
	public static final int VISIT_WALK = 1;
	public static final int VISIT_GRAZE = 2;

	// The proportion of red and blue to green which produces
	// the colour blend reflected by highest quality pastures 
	private static final float PROPORTION_RED = 0.45f, PROPORTION_BLUE = 0.5f;
	
	
	// Minimum number of milliseconds between re-filtering
	private static final int FILTER_INTERVAL = 5000;
	
	// Transparency masks	
	private static final int OPAQUE = 0xFF000000;
	private static final int ALPHA_MASK = 0xFF000000;
	
	// Colour masks
	private static final int RED_MASK   = 0x00FF0000;
	private static final int GREEN_MASK = 0x0000FF00;
	private static final int BLUE_MASK  = 0x000000FF;
	
	// Soil data masks
	private static final int MASK_NI  = 0x0000000F;
	private static final int MASK_SWC = 0x000000F0;
	private static final int MASK_AW  = 0x0000FF00;
	private static final int MASK_VI  = 0x00FF0000;
	
	// Plant data masks
	private static final int MASK_LAI = 0x00FFF000;
	private static final int MASK_DM  = 0x00000FFF;
	
	private transient Simulation sim;     // A reference to the simulation
	private transient Image nImg;         // The filtered image representing the soil nitrogen
	private transient Image cImg;         // The filtered image representing the soil water capacity
	private transient Image wImg;         // The filtered image representing the soil water
	private transient Image hImg;         // The filtered image representing plant height 
	private transient Image aImg;         // The filtered image representing plant LAI
	private transient Image vImg;         // The filtered image representing the visit frequency
	private transient BufferedImage soilData;  // The buffered image storing the soil data
	private transient BufferedImage plantData; // The buffered image storing the plant data
	private transient long lastFilter;    // Time since the last re-filter of pasture data

	// For stats
	private long LAIstats, DMstats, AWstats, OVRstats;

	
	private File soilImageFile;  // The file holding the siol base image
	private File plantImageFile; // The file holding the plant base image
	private int width, height;   // The dimensions of the image
	private int regenCount;      // The counter which keeps track of which area is being regenerated
	private int growthCellSize;  // The size of a regen square
	private GrassModel grass;    // The model governing the grass growth
	
	public double getRainfall()     { return grass.getRainfall();      }
	public double getTemperature()  { return grass.getTemperature();   }
	public double getRadiation()    { return grass.getRadiation();     }
	public double getTranspiration(){ return grass.getTranspiration(); }
	public void setRainfall(double r)     { grass.setRainfall(r);      }
	public void setTemperature(double t)  { grass.setTemperature(t);   }
	public void setRadiation(double r)    { grass.setRadiation(r);     }
	public void setTranspiration(double t){ grass.setTranspiration(t); }
	
	
	/** Constructs a new pasture state image from the image file */
	public PastureState(Simulation sim, File imgFile){
		this.sim = sim;
		
//Popup wait = new Popup("Loading aerial photograph.");

		BufferedImage data = sim.loadImage(imgFile, 1);
		this.width = data.getWidth(null);
		this.height = data.getHeight(null);
		regenCount = 0;
		growthCellSize = (int)Math.sqrt((((double)width*(double)height)/(60.0*60.0*24.0/(double)Simulation.TIME_FACTOR)))+2;
		grass = new GrassModel(sim);
				
		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice gs = ge.getDefaultScreenDevice();
		GraphicsConfiguration gc = gs.getDefaultConfiguration();
		soilData = gc.createCompatibleImage(width, height, BufferedImage.TYPE_INT_RGB);
		plantData = gc.createCompatibleImage(width, height, BufferedImage.TYPE_INT_RGB);
		
//wait.setMessage("Calibrating pasture...");
		
		for(int i = 0; i < width; i++){
			for(int j = 0; j < height; j++){
				int pixel = data.getRGB(i,j);
				int NI = estimateNI(pixel); // Estimate nitrogen from image
				int SWC = estimateSWC(pixel); // Estimate water capacity from image
//grass.setData(NI, SWC, 0, 0, 0);
//grass.prepare(); // Grow grass for a couple weeks
//soilData.setRGB( i, j, ((grass.getAW()<<8) & MASK_AW) | ((grass.getSWC()<<4) & MASK_SWC) | (grass.getNI() & MASK_NI) );
//plantData.setRGB( i, j, ((grass.getLAI()<<12) & MASK_LAI) | (grass.getDM() & MASK_DM) );
				soilData.setRGB( i, j, ((SWC<<4) & MASK_SWC) | (NI & MASK_NI) );
//plantData.setRGB( i, j, ((grass.getLAI()<<12) & MASK_LAI) | (grass.getDM() & MASK_DM) );
			}
//wait.setMessage("Calibrating pasture: "+(int)(i*100/width+0.5)+"% complete");
			System.out.print("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\bCALIBRATING GRASS: "+(int)(i*100/width+0.5)+"% COMPLETE");
		}
		System.out.print("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\bCALIBRATING GRASS: 100% COMPLETE");
		System.out.println("\nCALIBRATING GRASS DONE");
//wait.setMessage("Calibrating pasture: 100% complete");
		
		doFilter();
//wait.dispose();
	}

	/** Initialise the pasture data after a load */
	public void initialise(Simulation sim){ grass.initialise(sim); loadImage(sim); }
	
	
	/** Remove SWC for all areas over fences */
	public void newBounds(FieldBounds f){
		Rectangle b = f.getBounds();
		b.x = (b.x < 1)? 1 : b.x;
		b.y = (b.y < 1)? 1 : b.y;
		b.width = (b.x + b.width >= width)? width - b.x - 1 : b.width;
		b.height = (b.y + b.height >= height)? height - b.x - 1 : b.height;
		if(b.width < 3 || b.height < 3) return;
		for(int i = b.x-1; i < b.x+b.width+1; i++){
			for(int j = b.y-1; j < b.y+b.height+1; j++){
				if(
				!(f.contains(i-0.5, j-0.5) &&
				  f.contains(i-0.5, j+0.5) &&
				  f.contains(i+0.5, j-0.5) &&
				  f.contains(i+0.5, j+0.5))
				  &&
				!(!f.contains(i-0.5, j-0.5) &&
				  !f.contains(i-0.5, j+0.5) &&
				  !f.contains(i+0.5, j-0.5) &&
				  !f.contains(i+0.5, j+0.5))
				  ){
					int soil = soilData.getRGB(i,j);
					soilData.setRGB( i, j, soil & ~MASK_SWC);
				}
			}
		}
	}
	
	/** Remove SWC for all areas over water */
	public void newWater(FieldWater w){
		Rectangle b = w.getBounds();
		b.x = (b.x < 1)? 1 : b.x;
		b.y = (b.y < 1)? 1 : b.y;
		b.width = (b.x + b.width >= width)? width - b.x - 1 : b.width;
		b.height = (b.y + b.height >= height)? height - b.x - 1 : b.height;
		if(b.width < 3 || b.height < 3) return;
		for(int i = b.x-1; i < b.x+b.width+1; i++){
			for(int j = b.y-1; j < b.y+b.height+1; j++){
				if(w.contains(i-0.5, j-0.5) ||
				   w.contains(i-0.5, j+0.5) ||
				   w.contains(i+0.5, j-0.5) ||
				   w.contains(i+0.5, j+0.5)){
					int soil = soilData.getRGB(i,j);
					soilData.setRGB( i, j, soil & ~MASK_SWC);
				}
			}
		}
	}
	
	/** Restore SWC for all areas over fences */
	public void boundsRemoved(FieldBounds f){
		Rectangle b = f.getBounds();
		b.x = (b.x < 1)? 1 : b.x;
		b.y = (b.y < 1)? 1 : b.y;
		b.width = (b.x + b.width >= width)? width - b.x - 1 : b.width;
		b.height = (b.y + b.height >= height)? height - b.x - 1 : b.height;
		if(b.width < 3 || b.height < 3) return;
		for(int i = b.x-1; i < b.x+b.width+1; i++){
			for(int j = b.y-1; j < b.y+b.height+1; j++){
				if(
				!(f.contains(i-0.5, j-0.5) &&
				  f.contains(i-0.5, j+0.5) &&
				  f.contains(i+0.5, j-0.5) &&
				  f.contains(i+0.5, j+0.5))
				  &&
				!(!f.contains(i-0.5, j-0.5) &&
				  !f.contains(i-0.5, j+0.5) &&
				  !f.contains(i+0.5, j-0.5) &&
				  !f.contains(i+0.5, j+0.5))
				  ){
					int soil = soilData.getRGB(i,j);
					soilData.setRGB( i, j, soil | (MASK_SWC & ((soil & MASK_NI)<<4)));
				}
			}
		}
	}
	
	/** Restore SWC for all areas over water */
	public void waterRemoved(FieldWater w){
		Rectangle b = w.getBounds();
		b.x = (b.x < 1)? 1 : b.x;
		b.y = (b.y < 1)? 1 : b.y;
		b.width = (b.x + b.width >= width)? width - b.x - 1 : b.width;
		b.height = (b.y + b.height >= height)? height - b.x - 1 : b.height;
		if(b.width < 3 || b.height < 3) return;
		for(int i = b.x-1; i < b.x+b.width+1; i++){
			for(int j = b.y-1; j < b.y+b.height+1; j++){
				if(w.contains(i-0.5, j-0.5) ||
				   w.contains(i-0.5, j+0.5) ||
				   w.contains(i+0.5, j-0.5) ||
				   w.contains(i+0.5, j+0.5)){
					int soil = soilData.getRGB(i,j);
					soilData.setRGB( i, j, soil | (MASK_SWC & ((soil & MASK_NI)<<4)));
				}
			}
		}
	}
	
		
	public void loadImage(Simulation sim){
System.out.print("LOADING IMAGES...\n\t" + soilImageFile + "\n\t"+plantImageFile);
		this.sim = sim;
		try{
			//if(imageFile == null) imageFile = sim.nextImageFile();
			soilData = sim.loadImage(soilImageFile, 1);
			plantData = sim.loadImage(plantImageFile, 2);
//System.out.println("DATA = "+data);
			//data = ImageIO.read(imageFile);
System.out.println("\n ...LOAD SUCCESSFUL");
		}catch(Exception e){
System.out.println("\n ...LOAD FAILED");
			e.printStackTrace();
		}
		doFilter();
	}
	
	public void doFilter(){
		lastFilter = 0;
		filterN();
		lastFilter = 0;
		filterC();
		lastFilter = 0;
		filterW();
		lastFilter = 0;
		filterH();
		lastFilter = 0;
		filterA();
		lastFilter = 0;
		filterV();
	}
	public void filterN(){
		if(System.currentTimeMillis() - lastFilter > FILTER_INTERVAL){
			ImageFilter filter = new NitrogenFilter();
			nImg = Toolkit.getDefaultToolkit().createImage(new FilteredImageSource(soilData.getSource(), filter));
			lastFilter = System.currentTimeMillis();
		}
	}
	public void filterC(){
		if(System.currentTimeMillis() - lastFilter > FILTER_INTERVAL){
			ImageFilter filter = new CapacityFilter();
			cImg = Toolkit.getDefaultToolkit().createImage(new FilteredImageSource(soilData.getSource(), filter));
			lastFilter = System.currentTimeMillis();
		}
	}
	public void filterW(){
		if(System.currentTimeMillis() - lastFilter > FILTER_INTERVAL){
			ImageFilter filter = new WaterFilter();
			wImg = Toolkit.getDefaultToolkit().createImage(new FilteredImageSource(soilData.getSource(), filter));
			lastFilter = System.currentTimeMillis();
		}
	}
	public void filterH(){
		if(System.currentTimeMillis() - lastFilter > FILTER_INTERVAL){
			ImageFilter filter = new HeightFilter();
			hImg = Toolkit.getDefaultToolkit().createImage(new FilteredImageSource(plantData.getSource(), filter));
			lastFilter = System.currentTimeMillis();
		}
	}
	public void filterA(){
		if(System.currentTimeMillis() - lastFilter > FILTER_INTERVAL){
			ImageFilter filter = new LeafFilter();
			aImg = Toolkit.getDefaultToolkit().createImage(new FilteredImageSource(plantData.getSource(), filter));
			lastFilter = System.currentTimeMillis();
		}
	}
	public void filterV(){
		if(System.currentTimeMillis() - lastFilter > FILTER_INTERVAL){
			ImageFilter filter = new VisitFilter();
			vImg = Toolkit.getDefaultToolkit().createImage(new FilteredImageSource(soilData.getSource(), filter));
			lastFilter = System.currentTimeMillis();
		}
	}
	

	/** Estimates nitrogen index by scanning pixel values */	
	private static int estimateNI(int rgb) {
		
		int red = (rgb & RED_MASK) >> 16;
		int green = (rgb & GREEN_MASK) >> 8;
		int blue = rgb & BLUE_MASK;
		
		// Gives the brightness of the pixel between 0 and 1
		float brightness = ((float)red+green+blue)/(255*3);
		
		// Find how close the shade matches to perfect green grass
		int redOffset = Math.abs((int)(green - red/PROPORTION_RED));
		int blueOffset = Math.abs((int)(green - blue/PROPORTION_BLUE));
		
		// Gives a number between 0 and 1 representing the percentage
		// difference between this and an optimal green blend
		float missmatch = (redOffset + blueOffset) / (255/PROPORTION_RED + 255/PROPORTION_BLUE);

		//if(brightness < 0 || brightness > 1) System.out.println("missmatch overflow! ("+brightness+")");
		// Contruct the biomass density pixel
		//int bPixel;
		//if(missmatch > 0.4 || brightness < 0.1 || brightness > 0.8) bPixel = 0x00000000; 
		//else bPixel = (int)((1-brightness)*65535);
		
		// Construct the nutrition pixel
		//if(brightness < 0.05) nPixel = 0x00000000; // If really dark it is probably water 
		//else nPixel = (int)((1-(missmatch * brightness))*255);
		int nValue = (int)((1-brightness)*16);
				
		if(nValue < 0 || nValue > 16) System.out.println("nutrition overflow! ("+nValue+")");
		
		return nValue; 
	}
	
	
	/** Estimates nitrogen index by scanning pixel values */	
	private static int estimateSWC(int rgb) {
		
		int red = (rgb & RED_MASK) >> 16;
		int green = (rgb & GREEN_MASK) >> 8;
		int blue = rgb & BLUE_MASK;
		

		// Gives the brightness of the pixel between 0 and 1
		float brightness = ((float)red+green+blue)/(255*3);

		float missmatch = plantColourMissmatch(red, green, blue);
		
		// Calculate soil capacity
		int cValue;
		if(missmatch > 0.4 || brightness < 0.1 || brightness > 0.8) cValue = 0; 
		else cValue = (int)((1-brightness)*16);
		
		if(cValue < 0 || cValue > 65535) System.out.println("capacity overflow! ("+cValue+")");
		return cValue; 
	}

	
	/** Returns a float between 0 and 1 indicating the level
	    of missmatch between this colour and plant colour */
	private static float plantColourMissmatch(int r, int g, int b){
		
		// Find how close the shade matches to perfect green grass
		int redOffset = Math.abs((int)(g - r/PROPORTION_RED));
		int blueOffset = Math.abs((int)(g - b/PROPORTION_BLUE));
		
		// Gives a number between 0 and 1 representing the percentage
		// difference between this and an optimal green blend
		return (redOffset + blueOffset) / (255/PROPORTION_RED + 255/PROPORTION_BLUE);
	}
	

	
	/** Simulate pasture growth to replenish grass quality */
	public void regenerate(){
		
		int fullWidth = width + width % growthCellSize;
	
		int xp = regenCount % ((int)(fullWidth/growthCellSize));
		int yp = regenCount / ((int)(fullWidth/growthCellSize));
		//int yp = (int)((regenCount * growthCellSize) / fullWidth);
		int xlim, ylim;
		
		// Increase coordinates to correct cell
		xp *= growthCellSize; yp *= growthCellSize;
		
		// Set the limits to count up to, prevent out of bounds excxeption
		if(xp + growthCellSize >= width) xlim = width-1;
		else xlim = xp + growthCellSize;
		if(yp + growthCellSize >= height) ylim = height-1;
		else ylim = yp + growthCellSize;
		
		
		// Grow the pixels in the cell		
		for(int i = xp; i < xlim; i++){
			for(int j = yp; j < ylim; j++){
				int soil = soilData.getRGB(i,j);
				int plant = plantData.getRGB(i,j);
				grass.setData( soil&MASK_NI, (soil&MASK_SWC)>>4, (soil&MASK_AW)>>8, (plant&MASK_LAI)>>12, plant&MASK_DM );
				grass.simulate();
				// Visit data from last day is dampened by 1.5X
				int visit = (int)(((soil&MASK_VI)>>16)/1.5);
				
				// For Stats
				int LAIs = grass.getLAI();
				LAIstats += LAIs;
				OVRstats += (LAIs == 0)? 1 : 0;
				DMstats += grass.getDM();
				AWstats += grass.getAW();

													
				soilData.setRGB( i, j, ((visit<<16)&MASK_VI) | ((grass.getAW()<<8) & MASK_AW) | ((grass.getSWC()<<4) & MASK_SWC) | (grass.getNI() & MASK_NI) );
				plantData.setRGB( i, j, ((grass.getLAI()<<12) & MASK_LAI) | (grass.getDM() & MASK_DM) );
			}
		}
		
		regenCount += 1;
		
		// Reset regen count at the end of the day
		if(regenCount >= 60*24){ 
			regenCount = 0;
			
			long area = width*height;
			if(PastureSim.printPastureData){
				PastureSim.writePasture(str((int)((100l*(LAIstats/area))/4096l))+str((int)((100l*(DMstats/area))/4096l))+str((int)((100l*(AWstats/area))/256l))+str((int)((100l*OVRstats)/area))+"\n", sim.minute());
				System.out.println(""+sim.minute()+" LAI:"+str((int)((100l*(LAIstats/area))/4096l))+" DM:"+str((int)((100l*(DMstats/area))/4096l))+" AW:"+str((int)((100l*(AWstats/area))/256l))+" OVR:"+str((int)((100l*OVRstats)/area)));
			}
//System.out.println(""+sim.minute()+" LAI:"+str((int)((100l*(LAIstats/area))/4096l))+" DM:"+str((int)((100l*(DMstats/area))/4096l))+" AW:"+str((int)((100l*(AWstats/area))/256l))+" OVR:"+str((int)((100l*OVRstats)/area)));
			
			if(sim.pastureGraphData == null){
				sim.pastureGraphData = new GraphData[4];
				for(int i = 0; i < 4; i++){ sim.pastureGraphData[i] = new GraphData(); }
			}
			sim.pastureGraphData[0].updateDay(sim.minute(), (double)LAIstats/((double)area*4096.0));
//System.out.println("setting lai to"+((double)LAIstats/((double)area*4096.0))+" Lai = "+LAIstats);
			sim.pastureGraphData[1].updateDay(sim.minute(), (double)DMstats/((double)area*4096.0));
			sim.pastureGraphData[2].updateDay(sim.minute(), (double)AWstats/((double)area*256.0));
			sim.pastureGraphData[3].updateDay(sim.minute(), (double)OVRstats/(double)area);
				
			LAIstats = 0; DMstats = 0; AWstats = 0; OVRstats = 0;
		}
	}


	/** Convert int to percentage for printing*/
	public String str(int num){
		if(num < 10) return ""+num+"   ";
		if(num < 100) return ""+num+"  ";
		return ""+num+" ";
	}
	
	
	/** Simulates the visit to a pixel by an animal */	
	public float visit(int x, int y, int type){ return visit(x, y, type, 0, 0); }	
	public float visit(int x, int y, int type, float mass, int aheight){
		float energy = 0.0f;
	
		if(x >= 0 && y >= 0 && x < width && y < height){
			// Get the data
			int soil = soilData.getRGB(x,y);
			int plant = plantData.getRGB(x,y);
			grass.setData( soil&MASK_NI, (soil&MASK_SWC)>>4, (soil&MASK_AW)>>8, (plant&MASK_LAI)>>12, plant&MASK_DM );
			int visit = (soil&MASK_VI)>>16;
			
			// Do the update
			//if(pop < 255) pop++;
			if(type == VISIT_WALK) { energy = grass.trample(); visit = (visit < 253)? visit + 2: 255; }
			else if(type == VISIT_STAND){ energy = grass.trample(); }// visit = (visit < 255)? ++visit: 255; }
			else if(type == VISIT_GRAZE){ energy = grass.consume(mass, aheight); }//grass.trample(); visit = (visit < 245)? visit + 10: 255; }
			
			
			
			soilData.setRGB( x, y, ((visit<<16) & MASK_VI) | ((grass.getAW()<<8) & MASK_AW) | ((grass.getSWC()<<4) & MASK_SWC) | (grass.getNI() & MASK_NI) );
			plantData.setRGB( x, y, ((grass.getLAI()<<12) & MASK_LAI) | (grass.getDM() & MASK_DM) );
		}
		
		return energy;
	}
	
	// Returns the value of this pixel
	public float value(int x, int y, int aheight){
		if(x >= 0 && y >= 0 && x < width && y < height){
			return findVal(x, y,  aheight);
		}
		return 0.0f;
	}
	
	// Returns the average value of five pixels around the point
	public float avgVal(int x, int y, int aheight){
		if(x >= 0 && y >= 0 && x < width && y < height){
			float vC =                 findVal(x, y,  aheight);
			float vN = (y > 0)?        findVal(x, y-1,aheight) : 0.0f;
			float vS = (y < height-1)? findVal(x, y+1,aheight) : 0.0f;
			float vE = (x > 0)?        findVal(x-1, y,aheight) : 0.0f;
			float vW = (x < width-1)?  findVal(x+1, y,aheight) : 0.0f;
			return vC + vN + vS + vE + vW;
		}
		return 0.0f;
	}
	
	private float findVal(int x, int y, int aheight){
		int soil = soilData.getRGB(x,y);
		int plant = plantData.getRGB(x,y);
		grass.setData( soil&MASK_NI, (soil&MASK_SWC)>>4, (soil&MASK_AW)>>8, (plant&MASK_LAI)>>12, plant&MASK_DM );
		return grass.energyValue(aheight);
	}
	
	
	/** Returns visit frequency for that pixel */
	public int freq(int x, int y){
		if(x >= 0 && y >= 0 && x < width && y < height){
			int freqC = (soilData.getRGB(x,y)&MASK_VI)>>16;
			int freqN = (y > 0)?        (soilData.getRGB(x,y-1)&MASK_VI)>>16 : 0;
			int freqS = (y < height-1)?  (soilData.getRGB(x,y+1)&MASK_VI)>>16 : 0;
			int freqE = (x > 0)?        (soilData.getRGB(x-1,y)&MASK_VI)>>16 : 0;
			int freqW = (x < width-1)? (soilData.getRGB(x+1,y)&MASK_VI)>>16 : 0;
			
			return freqC + freqN + freqE + freqS + freqW;
		}
		return 0;
	}

	
	// Draws the filtered data images
	
	public void drawNitrogen(Graphics g, int x, int y, int width, int height){
		g.drawImage(nImg, x, y, width, height, null);
	}
	public void drawCapacity(Graphics g, int x, int y, int width, int height){
		g.drawImage(cImg, x, y, width, height, null);
	}
	public void drawWater(Graphics g, int x, int y, int width, int height){
		g.drawImage(wImg, x, y, width, height, null);
	}
	public void drawGrassHeight(Graphics g, int x, int y, int width, int height){
		g.drawImage(hImg, x, y, width, height, null);
	}
	public void drawLeafArea(Graphics g, int x, int y, int width, int height){
		g.drawImage(aImg, x, y, width, height, null);
	}
	public void drawVisitFreq(Graphics g, int x, int y, int width, int height){
		g.drawImage(vImg, x, y, width, height, null);
	}

	
	
	// Image filters for data

	class NitrogenFilter extends RGBImageFilter{
		public int filterRGB(int x, int y, int rgb) {
			int NI = (rgb&MASK_NI) << 4; // Multiply by 16
			return (OPAQUE) | ((255-NI)<<16) | ((128-(int)(NI/2))<<8);
		}
	}
	class CapacityFilter extends RGBImageFilter{
		public int filterRGB(int x, int y, int rgb) {
			int SWC = (rgb&MASK_SWC); // Already multiplied by 16
			return (OPAQUE) | ((255-SWC)<<16) | (SWC);
		}
	}
	class WaterFilter extends RGBImageFilter{
		public int filterRGB(int x, int y, int rgb) {
			int AW = (rgb&MASK_AW)>>8;
			return (OPAQUE) | ((255-AW)<<16) | ((255-AW)<<8) | 255;
		}
	}

	/*	
	An attempt to combine LAI and Grass height to view overall pasture
	quality but it ended up being a better idea to do them seperately
	
	class PlantFilter extends RGBImageFilter{
		public int filterRGB(int x, int y, int rgb){
			int LAI = 220 - (int)(((rgb & MASK_LAI)>>12)*220/4096); //(220 to 0)
			int DM = 255 - (int)((rgb & MASK_DM)/16); //(255 to 0)			
			//System.out.print(" LAI:"+LAI);
			//LAI = (LAI >= 0)? 0 : LAI;
			//if(LAI < 0){ System.out.print(" LAI:"+LAI); LAI = 0; }
			
			// Old scheme			
			//int LAI = 255 - (int)(((rgb & MASK_LAI)>>12)/16); //(255 to 0)
			//float DM = ((float)(rgb & MASK_DM)/(32768.0f))+1.0f; //(1 to 1.5)			
			//int red = (int)(((float)LAI)/DM);
			//int green = (int)(255.0f/DM);
			
			return (OPAQUE) | (LAI<<16) | (255<<8) | DM;
		}
	}
	*/
	
	class HeightFilter extends RGBImageFilter{
		public int filterRGB(int x, int y, int rgb){
			int DM = (int)((double)(rgb & MASK_DM)*200.0/4096.0); //(0 to 200)
			int red = 200 - DM;
			int green = (int)(red/4)+100;
			return (OPAQUE) | (red<<16) | (green<<8) | 0;
		}
	}
	class LeafFilter extends RGBImageFilter{
		public int filterRGB(int x, int y, int rgb){
			double LAI = Math.pow(((double)((rgb & MASK_LAI)>>12)/4095.0), 0.5); //(0 to 4095)
			if(LAI == 0.0) return (OPAQUE) | 0x00000000; // return black for none
			int val = (int)(LAI * 255);
			return (OPAQUE) | ((255-val)<<16) | (val<<8);
		}
	}
	
	class VisitFilter extends RGBImageFilter{
		public int filterRGB(int x, int y, int rgb){
			int freq = (rgb & MASK_VI) >> 16; // (0 to 255)
			return (OPAQUE) | (freq<<16) | (255-freq);
		}
	}
	
	
	private void writeObject(java.io.ObjectOutputStream out) throws IOException{
		soilImageFile = sim.nextSoilImageFile();
		plantImageFile = sim.nextPlantImageFile();
		out.writeInt(width);
		out.writeInt(height);
		out.writeInt(regenCount);
		out.writeInt(growthCellSize);
		out.writeObject(grass);
		out.writeLong(LAIstats);
		out.writeLong(DMstats);
		out.writeLong(AWstats);
		out.writeLong(OVRstats);
System.out.print("SAVING... "+soilImageFile);
		out.writeObject(soilImageFile); ImageIO.write(soilData, "png", soilImageFile); 
System.out.println(" ...DONE");

System.out.print("SAVING... "+plantImageFile);
		out.writeObject(plantImageFile); ImageIO.write(plantData, "png", plantImageFile); 
System.out.println(" ...DONE");

	}
	
	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException{
		width = in.readInt();
		height = in.readInt();
		regenCount = in.readInt();
		growthCellSize = in.readInt();
		grass = (GrassModel)in.readObject();
		LAIstats = in.readLong();
		DMstats = in.readLong();
		AWstats = in.readLong();
		OVRstats = in.readLong();
		soilImageFile = (File)in.readObject();
		plantImageFile = (File)in.readObject();
	}
	
}
