/*
 * File:   PathFinderMap.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */


package utils;

import gui.*;
import sim.*;
import sim.pasture.*;
import java.io.*;
import java.util.*;
import javax.swing.JOptionPane;
import java.awt.*;
import java.awt.geom.*;


/** Creates paths to water using a variant of the A* path finding algorithm */
public class PathFinderMap implements Serializable{

	
	// The number of meters a cell occupies
	public static final float RELATIVE_CELL_SIZE = (float)Simulation.ACTUAL_GATE_WIDTH/2.0f;
	
	public static final float ROOT_TWO = (float)Math.sqrt(2.0);
	
	// The size of each cell in the map
	private int MAP_CELL_SIZE;
	
	// The max distance
	private int MAX_DIST;
	private int MAX_CELL_DIST;

	// The direction constants
	private static final byte NE = 1;
	private static final byte SE = 2;
	private static final byte SW = 3;
	private static final byte NW = 4;
	private static final byte N  = 5;
	private static final byte E  = 6;
	private static final byte S  = 7;
	private static final byte W  = 8;
	private static final byte WATER = 9;
	
	
	private Simulation sim;   // A reference to the simulation
	private byte[][] map;     // The path map
	private float[][] dist;   // The distance array
	private Dimension d;      // the size of the area
	private LinkedList water; // All water in the pasture

	
	/** Construct a dummy path map for a new simulation */
	public PathFinderMap(){
		sim = null;
		map = null;
		dist = null;
		d = null;
		water = null;
	}
	
	/** Construct a path map for a simulation with water */
	public PathFinderMap(Simulation sim){
		this.sim = sim;
		water = sim.getWaterGeometry().water;
		d = sim.getPastureSize();
		if(d.width > d.height) MAX_DIST = (int)((double)d.width/sim.METER);
		else MAX_DIST = (int)((double)d.height/sim.METER);
		MAP_CELL_SIZE = (int)(sim.METER * RELATIVE_CELL_SIZE);
		if(MAP_CELL_SIZE == 0) MAP_CELL_SIZE = 1;
				
		// Initialise basic variables;
		int cols = (d.width + MAP_CELL_SIZE -1)/MAP_CELL_SIZE;
		int rows = (d.height + MAP_CELL_SIZE -1)/MAP_CELL_SIZE;
		if(rows > cols) MAX_CELL_DIST = rows;
		else MAX_CELL_DIST = cols;
		map = new byte[cols][rows];
		dist = initDist(cols, rows);
		// Construct the map
		if(water.size() > 0) growPath(findWater());
	}
	
	
	// Initialises distance array
	public float[][] initDist(int c, int r){
		float[][] dist = new float[c][r];
		for(int i = 0; i < c; i++){
			for(int j = 0; j < r; j++){
				dist[i][j] = -1.0f;//MAX_CELL_DIST-1;
			}
		}
		return dist;
	}
	
	
	// Returns a list of all the cells with water, as well as updating the
	// map array to contain the water
	private TreeSet findWater(){
		// Fill the map with water
		TreeSet waterCells = new TreeSet();
		for(int i = 0; i < map.length; i++){
			for(int j = 0; j < map[0].length; j++){
				// The centre of this cell
				Point p = findPoint(i, j);
				for(Iterator iter = water.iterator(); iter.hasNext();){
					FieldWater fw = (FieldWater)iter.next();
					// If this point is in the water then update map
					if(fw.contains(centrePoint(p))){
						AnimalContainer ac = sim.getFieldGeometry().findContainingField(p, MAP_CELL_SIZE);
						if(ac != null){
							map[i][j] = WATER;
							dist[i][j] = 0.0f;
							waterCells.add(new WaterSource(new Point(i,j), ac, 0.0f));
						}
					}
				}
			}
		}
		return waterCells;
	}
	
	
	// Grows a path matrix to move towards all cells in v1
	private void growPath(TreeSet v1){
		int cols = map.length; int rows = map[0].length;
		while(v1.size() > 0){
			
			TreeSet v2 = new TreeSet();		
			// Grow direction matrix for all adjacent cells
			for(Iterator i = v1.iterator(); i.hasNext(); ){
				WaterSource ws = (WaterSource)i.next();
				Point p = ws.getPos();
				AnimalContainer ac = ws.getContainer();
				if(withinBounds(p.x-1,p.y) && available(p.x-1,p.y,ac,v2,dist[p.x][p.y]+1.0f,E)){ map[p.x-1][p.y] = E; dist[p.x-1][p.y] = dist[p.x][p.y]+1.0f; v2.add(new WaterSource(new Point(p.x-1, p.y),ac,dist[p.x][p.y]+1.0f)); }
				if(withinBounds(p.x+1,p.y) && available(p.x+1,p.y,ac,v2,dist[p.x][p.y]+1.0f,W)){ map[p.x+1][p.y] = W; dist[p.x+1][p.y] = dist[p.x][p.y]+1.0f; v2.add(new WaterSource(new Point(p.x+1, p.y),ac,dist[p.x][p.y]+1.0f)); }
				if(withinBounds(p.x,p.y-1) && available(p.x,p.y-1,ac,v2,dist[p.x][p.y]+1.0f,S)){ map[p.x][p.y-1] = S; dist[p.x][p.y-1] = dist[p.x][p.y]+1.0f; v2.add(new WaterSource(new Point(p.x, p.y-1),ac,dist[p.x][p.y]+1.0f)); }
				if(withinBounds(p.x,p.y+1) && available(p.x,p.y+1,ac,v2,dist[p.x][p.y]+1.0f,N)){ map[p.x][p.y+1] = N; dist[p.x][p.y+1] = dist[p.x][p.y]+1.0f; v2.add(new WaterSource(new Point(p.x, p.y+1),ac,dist[p.x][p.y]+1.0f)); }
			}
			// Grow direction matrix for all diagonal cells
			for(Iterator i = v1.iterator(); i.hasNext(); ){
				WaterSource ws = (WaterSource)i.next();
				Point p = ws.getPos();
				AnimalContainer ac = ws.getContainer();
				if(withinBounds(p.x-1,p.y-1) && available(p.x-1,p.y-1,ac,v2,dist[p.x][p.y]+ROOT_TWO,SE)){ map[p.x-1][p.y-1] = SE; dist[p.x-1][p.y-1] = dist[p.x][p.y]+ROOT_TWO; v2.add(new WaterSource(new Point(p.x-1, p.y-1),ac,dist[p.x][p.y]+ROOT_TWO)); }
				if(withinBounds(p.x-1,p.y+1) && available(p.x-1,p.y+1,ac,v2,dist[p.x][p.y]+ROOT_TWO,NE)){ map[p.x-1][p.y+1] = NE; dist[p.x-1][p.y+1] = dist[p.x][p.y]+ROOT_TWO; v2.add(new WaterSource(new Point(p.x-1, p.y+1),ac,dist[p.x][p.y]+ROOT_TWO)); }
				if(withinBounds(p.x+1,p.y-1) && available(p.x+1,p.y-1,ac,v2,dist[p.x][p.y]+ROOT_TWO,SW)){ map[p.x+1][p.y-1] = SW; dist[p.x+1][p.y-1] = dist[p.x][p.y]+ROOT_TWO; v2.add(new WaterSource(new Point(p.x+1, p.y-1),ac,dist[p.x][p.y]+ROOT_TWO)); }
				if(withinBounds(p.x+1,p.y+1) && available(p.x+1,p.y+1,ac,v2,dist[p.x][p.y]+ROOT_TWO,NW)){ map[p.x+1][p.y+1] = NW; dist[p.x+1][p.y+1] = dist[p.x][p.y]+ROOT_TWO; v2.add(new WaterSource(new Point(p.x+1, p.y+1),ac,dist[p.x][p.y]+ROOT_TWO)); }
// Old alg
//				if(withinBounds(p.x-1,p.y-1) && available(p.x-1,p.y-1,ac,v2,dist[p.x][p.y]+1.0f,SE)){ map[p.x-1][p.y-1] = SE; dist[p.x-1][p.y-1] = dist[p.x][p.y]+1.0f; v2.add(new WaterSource(new Point(p.x-1, p.y-1),ac,dist[p.x][p.y]+1.0f)); }
//				if(withinBounds(p.x-1,p.y+1) && available(p.x-1,p.y+1,ac,v2,dist[p.x][p.y]+1.0f,NE)){ map[p.x-1][p.y+1] = NE; dist[p.x-1][p.y+1] = dist[p.x][p.y]+1.0f; v2.add(new WaterSource(new Point(p.x-1, p.y+1),ac,dist[p.x][p.y]+1.0f)); }
//				if(withinBounds(p.x+1,p.y-1) && available(p.x+1,p.y-1,ac,v2,dist[p.x][p.y]+1.0f,SW)){ map[p.x+1][p.y-1] = SW; dist[p.x+1][p.y-1] = dist[p.x][p.y]+1.0f; v2.add(new WaterSource(new Point(p.x+1, p.y-1),ac,dist[p.x][p.y]+1.0f)); }
//				if(withinBounds(p.x+1,p.y+1) && available(p.x+1,p.y+1,ac,v2,dist[p.x][p.y]+1.0f,NW)){ map[p.x+1][p.y+1] = NW; dist[p.x+1][p.y+1] = dist[p.x][p.y]+1.0f; v2.add(new WaterSource(new Point(p.x+1, p.y+1),ac,dist[p.x][p.y]+1.0f)); }
			}
			// Execute for next furthest cells
			v1 = v2;
		}
	}
	
	// Checks wheather x and y are within the array bounds
	private boolean withinBounds(int x, int y){
		return x >= 0 && x < map.length && y > 0 && y < map[0].length;	
	}
	
	// Returns wheather the cell is available for this map
	private boolean available(int x, int y, AnimalContainer ac, TreeSet v2, float hop, byte direction){
		Point p = findPoint(x,y);
		
		// If it is within the area then return true
		if((map[x][y] == 0 || dist[x][y] > hop + ((Math.random()-0.5f)*0.00001f))  && ac.ownsArea(new Point2D.Double(p.x, p.y), MAP_CELL_SIZE)) return true;
// old alg
//		if((map[x][y] == 0 || dist[x][y] > hop)  && ac.ownsArea(new Point2D.Double(p.x, p.y), MAP_CELL_SIZE)) return true;
		
		// Check if any neighbours own the point,
		// if so, add them to the list to be processed
		AnimalContainer n = ac.findOpenNeighbours(centrePoint(p));
		if(n != null && map[x][y] == 0){
			map[x][y] = direction;
			dist[x][y] = hop;
			v2.add(new WaterSource(new Point(x,y), n, hop));
		}
		return false;
	}
	
	
	
	// Returns the (x,y) coord at the top left of the cell
	private Point findPoint(Point p){ return findPoint(p.x, p.y); }
	private Point findPoint(int x, int y){
		return new Point((x*MAP_CELL_SIZE), (y*MAP_CELL_SIZE));
	}
	
	// Returns the (x,y) coord at the centre center the cell
	private Point2D.Double centrePoint(Point p){ return centrePoint(p.x, p.y); }
	private Point2D.Double centrePoint(int x, int y){
		//return new Point(x, y);
		return new Point2D.Double(x+MAP_CELL_SIZE/2.0, y+MAP_CELL_SIZE/2.0);
	}
	
	// Returns the coordinates of the cell corresponding to the (x,y) coord
	private Point findCell(Point p){ return findCell(p.x, p.y); }
	private Point findCell(int x, int y){
		return new Point(x/MAP_CELL_SIZE, y/MAP_CELL_SIZE);
	}
	
	
	// Returns the distance of the nearest water in meters
	public float getDistance(double x, double y){ return getDistance((int)x, (int)y); }
	public float getDistance(int x, int y){
		if(sim == null || !sim.withinPastureBounds(x,y)) return -1.0f; // No water in proximity
		Point cell = findCell(x, y);
		float d = dist[cell.x][cell.y]*MAP_CELL_SIZE / (float)sim.METER; 
		return d;
	}
	
	
	
	// returns the direction vector for the direction to the nearest water
	public Vector2D getDirection(double x, double y){ return getDirection((int)x, (int)y); }
	public Vector2D getDirection(int x, int y){
		if(sim == null || !sim.withinPastureBounds(x,y)) return new Vector2D(0,0); // If dummy just return no direction
		Point cell = findCell(x, y);
		// Return correct direction vector
		switch(map[cell.x][cell.y]){
			case NE : return new Vector2D( 1,-1);
			case SE : return new Vector2D( 1, 1);
			case SW : return new Vector2D(-1, 1);
			case NW : return new Vector2D(-1,-1);
			case N  : return new Vector2D( 0,-1);
			case E  : return new Vector2D( 1, 0);
			case S  : return new Vector2D( 0, 1);
			case W  : return new Vector2D(-1, 0);
		}
		// Default to a random direction
		return new Vector2D();
	}
	
	
	// Returns wheather the polygon will be detected by this water map
	public static boolean willBeDetected(Simulation sim, Polygon area){
		// Get pasture information
		Dimension d = sim.getPastureSize();
		int MAP_CELL_SIZE = (int)(sim.METER * RELATIVE_CELL_SIZE);
		if(MAP_CELL_SIZE == 0) MAP_CELL_SIZE = 1;
		
		// get row and column size
		int cols = (d.width + MAP_CELL_SIZE -1)/MAP_CELL_SIZE;
		int rows = (d.height + MAP_CELL_SIZE -1)/MAP_CELL_SIZE;
		
		for(int i = 0; i < cols; i++){
			for(int j = 0; j < rows; j++){
				Point p = new Point((i*MAP_CELL_SIZE)-MAP_CELL_SIZE/2, (j*MAP_CELL_SIZE)-MAP_CELL_SIZE/2);
				if(area.contains(p)) return true;
			}
		}
		return false;
	}

	
	
	
	// Draw the data
	public void draw(Graphics g){
		if(sim == null) return; // No water paths to draw
		//drawDistance(g);		
		//drawDirection(g);
		//drawWaterSources(g);
	}

	// Draws distances
	public void drawDistance(Graphics g){
		if(sim == null) return; // No water paths to draw
		for(int i = 0; i < map.length; i++){
			for(int j = 0; j < map[0].length; j++){
				Point p = findPoint(i, j);
				float dist = (getDistance(p.x, p.y)/MAX_DIST);
				if(dist > 1.0) dist = 1.0f;
				if(dist > 0.0){
					g.setColor(new Color(dist,0.0f,1.0f-dist));
					int halfCell = (int)(MAP_CELL_SIZE/2);
					sim.fillScaledRect(g, p.x, p.y, MAP_CELL_SIZE, MAP_CELL_SIZE);
				}
			}
		}
	}
	
	// Draws directional arrows
	public void drawDirection(Graphics g){
		if(sim == null) return; // No water paths to draw
		g.setColor(Color.WHITE);
		double halfCell = MAP_CELL_SIZE/2.0f;
		double miniCell = MAP_CELL_SIZE/4.0f;
		for(int i = 0; i < map.length; i++){
			for(int j = 0; j < map[0].length; j++){
				Point2D.Double p = centrePoint(findPoint(i, j));
				double cx = p.x-miniCell;
				double cy = p.y-miniCell;
				switch(map[i][j]){
					case NE : sim.drawScaledLine(g, p.x, p.y, p.x+halfCell, p.y-halfCell); sim.drawScaledLine(g, p.x+halfCell, p.y-halfCell, p.x+halfCell-miniCell, p.y-halfCell); sim.drawScaledLine(g, p.x+halfCell, p.y-halfCell, p.x+halfCell, p.y-halfCell+miniCell); break; 
					case SE : sim.drawScaledLine(g, p.x, p.y, p.x+halfCell, p.y+halfCell); sim.drawScaledLine(g, p.x+halfCell, p.y+halfCell, p.x+halfCell-miniCell, p.y+halfCell); sim.drawScaledLine(g, p.x+halfCell, p.y+halfCell, p.x+halfCell, p.y+halfCell-miniCell); break;
					case SW : sim.drawScaledLine(g, p.x, p.y, p.x-halfCell, p.y+halfCell); sim.drawScaledLine(g, p.x-halfCell, p.y+halfCell, p.x-halfCell+miniCell, p.y+halfCell); sim.drawScaledLine(g, p.x-halfCell, p.y+halfCell, p.x-halfCell, p.y+halfCell-miniCell); break;
					case NW : sim.drawScaledLine(g, p.x, p.y, p.x-halfCell, p.y-halfCell); sim.drawScaledLine(g, p.x-halfCell, p.y-halfCell, p.x-halfCell+miniCell, p.y-halfCell); sim.drawScaledLine(g, p.x-halfCell, p.y-halfCell, p.x-halfCell, p.y-halfCell+miniCell); break;
					case N  : sim.drawScaledLine(g, p.x, p.y, p.x,          p.y-halfCell); sim.drawScaledLine(g, p.x         , p.y-halfCell, p.x-miniCell, p.y-halfCell+miniCell); sim.drawScaledLine(g, p.x         , p.y-halfCell, p.x+miniCell, p.y-halfCell+miniCell); break;
					case E  : sim.drawScaledLine(g, p.x, p.y, p.x+halfCell, p.y         ); sim.drawScaledLine(g, p.x+halfCell, p.y         , p.x+halfCell-miniCell, p.y+miniCell); sim.drawScaledLine(g, p.x+halfCell, p.y         , p.x+halfCell-miniCell, p.y-miniCell); break;
					case S  : sim.drawScaledLine(g, p.x, p.y, p.x,          p.y+halfCell); sim.drawScaledLine(g, p.x         , p.y+halfCell, p.x+miniCell, p.y+halfCell-miniCell); sim.drawScaledLine(g, p.x         , p.y+halfCell, p.x-miniCell, p.y+halfCell-miniCell); break;
					case W  : sim.drawScaledLine(g, p.x, p.y, p.x-halfCell, p.y         ); sim.drawScaledLine(g, p.x-halfCell, p.y         , p.x-halfCell+miniCell, p.y+miniCell); sim.drawScaledLine(g, p.x-halfCell, p.y         , p.x-halfCell+miniCell, p.y-miniCell); break;
				}
				
			}
		}
	}
		
	// Draw it ASCII art style!
	public String toString(){
		if(sim == null) return ""; // No water paths to draw
		String s = "";
		for(int j = 0; j < map[0].length; j++){
			for(int i = 0; i < map.length; i++){
				Point p = findPoint(i, j);
				switch(map[i][j]){
					case NE : s += "/"; break; 
					case SE : s += "\\"; break;
					case SW : s += "/"; break;
					case NW : s += "\\"; break;
					case N  : s += "|"; break;
					case E  : s += "-"; break;
					case S  : s += "|"; break;
					case W  : s += "-"; break;
					case WATER : s += "0"; break;
					default: s += " ";
				}
			}
			s += "\n";
		}
		
		return s;
	}
}
